###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=import-outside-toplevel, protected-access
import re
from importlib import reload
from os import environ, rename, getenv, path
from pathlib import Path
from sys import platform
from time import sleep
from everysk import config
from everysk.core.exceptions import FieldValueError
from everysk.core.threads import Thread
from everysk.core.unittests import TestCase, skipUnless, mock


class SettingsModulesTestCase(TestCase):

    def setUp(self) -> None:
        if platform == 'linux':
            self.root = '/var/app'
        elif platform == 'win32':
            self.root = 'C:\\var\\app'

        self.old_project_root = environ.get('PROJECT_ROOT', Undefined)
        self.old_everysk_root = config.EVERYSK_ROOT

    def tearDown(self) -> None:
        if self.old_project_root is not Undefined:
            environ['PROJECT_ROOT'] = self.old_project_root
        elif 'PROJECT_ROOT' in environ:
            environ.pop('PROJECT_ROOT')

        config.EVERYSK_ROOT = self.old_everysk_root

    @skipUnless(platform == 'linux', 'Runs only in Linux system.')
    def test_linux_simple_path(self):
        entry = Path('/var/app/settings.py')
        module = config.get_full_dotted_path_module(entry, self.root)
        self.assertEqual(module, 'settings')

    @skipUnless(platform == 'win32', 'Runs only in Windows system.')
    def test_windows_simple_path(self):
        entry = Path('C:\\var\\app\\settings.py')
        module = config.get_full_dotted_path_module(entry, self.root)
        self.assertEqual(module, 'settings')

    @skipUnless(platform == 'linux', 'Runs only in Linux system.')
    def test_linux_complex_path(self):
        entry = Path('/var/app/src/everysk/sdk/entities/settings.py')
        module = config.get_full_dotted_path_module(entry, self.root)
        self.assertEqual(module, 'src.everysk.sdk.entities.settings')

    @skipUnless(platform == 'win32', 'Runs only in Windows system.')
    def test_windows_complex_path(self):
        entry = Path('C:\\var\\app\\src\\everysk\\sdk\\entities\\settings.py')
        module = config.get_full_dotted_path_module(entry, self.root)
        self.assertEqual(module, 'src.everysk.sdk.entities.settings')

    def test_get_all_modules_everysk_lib(self):
        if 'PROJECT_ROOT' in environ:
            environ.pop('PROJECT_ROOT')

        modules = config.get_all_modules()
        # We could no guarantee the order of this list
        # the only thing that is needed is 'everysk.settings' to be the first
        self.assertEqual(modules[0], 'everysk.settings')

        # Now we check that every settings was read
        self.assertEqual(sorted(modules), [
            'everysk.sdk.engines.settings',
            'everysk.sdk.entities.custom_index.settings',
            'everysk.sdk.entities.datastore.settings',
            'everysk.sdk.entities.file.settings',
            'everysk.sdk.entities.portfolio.settings',
            'everysk.sdk.entities.private_security.settings',
            'everysk.sdk.entities.report.settings',
            'everysk.sdk.entities.settings',
            'everysk.sdk.entities.worker_execution.settings',
            'everysk.sdk.entities.workflow_execution.settings',
            'everysk.sdk.entities.workspace.settings',
            'everysk.sdk.settings',
            'everysk.server.settings',
            'everysk.settings'
        ])

    @skipUnless(path.exists('/var/app/test_project_root'), 'Project root does not exists.')
    def test_get_all_modules_project_root(self):
        environ['PROJECT_ROOT'] = '/var/app/test_project_root'
        config.EVERYSK_ROOT = ''
        modules = config.get_all_modules()
        self.assertListEqual(modules, ['everysk.settings'])

    def test_is_valid_path(self):
        self.assertTrue(config._is_valid_path('everysk/settings.py'))
        self.assertFalse(config._is_valid_path('site-packages/settings.py'))
        self.assertFalse(config._is_valid_path('.git/settings.py'))


class SettingsTestCase(TestCase):
    """
    Because we need the correct project root folder we import the class/functions inside every test
    """
    # we need to rename these files to avoid use then on prod environ
    old_settings = [f'{config.EVERYSK_ROOT}/core/fixtures/other/_settings.py', f'{config.EVERYSK_ROOT}/core/fixtures/_settings.py']
    new_settings = [f'{config.EVERYSK_ROOT}/core/fixtures/other/settings.py', f'{config.EVERYSK_ROOT}/core/fixtures/settings.py']

    @classmethod
    def setUpClass(cls) -> None:
        rename(cls.old_settings[0], cls.new_settings[0])
        rename(cls.old_settings[1], cls.new_settings[1])
        # Because we changed the settings files we need to reload the module
        del config.settings
        reload(config)

    @classmethod
    def tearDownClass(cls) -> None:
        rename(cls.new_settings[0], cls.old_settings[0])
        rename(cls.new_settings[1], cls.old_settings[1])
        # Because we changed the settings files we need to reload the module
        del config.settings
        reload(config)

    def test_get_all_settings_singleton(self):
        settings01 = config.Settings()
        settings02 = config.Settings()
        self.assertEqual(settings01, settings02)

    def test_not_singleton(self):
        settings01 = config.Settings(singleton=False)
        settings02 = config.Settings(singleton=False)
        self.assertNotEqual(settings01, settings02)

    def test_settings(self):
        self.assertEqual(config.settings.EVERYSK_TEST_NAME, 'test-case')
        self.assertEqual(config.settings.EVERYSK_TEST_OTHER_NAME, 'test-other-case')
        self.assertEqual(config.settings.EVERYSK_TEST_INT, 1)
        self.assertEqual(config.settings.EVERYSK_TEST_OTHER_INT, 2)
        self.assertTrue(config.settings.EVERYSK_TEST_FAKE)
        self.assertTrue(config.settings.EVERYSK_TEST_OTHER_FAKE)

        # See if the setting was replaced
        self.assertEqual(config.settings.EVERYSK_TEST_FIELD_INHERIT, 'test-case as True')
        self.assertEqual(config.settings.EVERYSK_TEST_VAR_INHERIT, 'True as test-case')
        self.assertEqual(config.settings.EVERYSK_TEST_OTHER_FIELD_INHERIT, 'test-case as True')
        self.assertEqual(config.settings.EVERYSK_TEST_OTHER_VAR_INHERIT, 'True as test-case')

        # Test readonly attributes
        with self.assertRaisesRegex(FieldValueError, "The field 'EVERYSK_TEST_NAME' value cannot be changed."):
            config.settings.EVERYSK_TEST_NAME = 'new-test'

        # Test validation
        with self.assertRaisesRegex(FieldValueError, "Key EVERYSK_TEST_INT must be <class 'int'>."):
            config.settings.EVERYSK_TEST_INT = 'new-test'

    def test_settings_environment_value(self):
        self.assertEqual(config.settings.EVERYSK_TEST_BOOL_VAR, False)
        self.assertEqual(config.settings.EVERYSK_TEST_FAKE, True)
        self.assertEqual(config.settings.EVERYSK_TEST_INT, 1)
        environ['EVERYSK_TEST_FAKE'] = '0'
        environ['EVERYSK_TEST_BOOL_VAR'] = '1'
        environ['EVERYSK_TEST_INT'] = '2'
        del config.settings
        reload(config)
        self.assertEqual(config.settings.EVERYSK_TEST_BOOL_VAR, True)
        self.assertEqual(config.settings.EVERYSK_TEST_FAKE, False)
        self.assertEqual(config.settings.EVERYSK_TEST_INT, 2)
        # Cleanup environ
        del environ['EVERYSK_TEST_FAKE']
        del environ['EVERYSK_TEST_BOOL_VAR']
        del environ['EVERYSK_TEST_INT']

    def test_config_pyi_file(self):
        config_pyi_file = None
        with open(f'{config.EVERYSK_ROOT}/config.pyi', 'r', encoding='utf-8') as arq:
            config_pyi_file = arq.read()

        self.assertIn('EVERYSK_TEST_OTHER_FAKE: bool', config_pyi_file)
        self.assertIn('EVERYSK_TEST_OTHER_FIELD_INHERIT: str', config_pyi_file)
        self.assertIn('EVERYSK_TEST_OTHER_INT: int', config_pyi_file)
        self.assertIn('EVERYSK_TEST_OTHER_NAME: str', config_pyi_file)
        self.assertIn('EVERYSK_TEST_OTHER_VAR_INHERIT: str', config_pyi_file)
        self.assertIn('EVERYSK_TEST_BOOL_VAR: bool', config_pyi_file)
        self.assertIn('EVERYSK_TEST_FAKE: bool', config_pyi_file)
        self.assertIn('EVERYSK_TEST_FIELD_INHERIT: str', config_pyi_file)
        self.assertIn('EVERYSK_TEST_INT: int', config_pyi_file)
        self.assertIn('EVERYSK_TEST_NAME: str', config_pyi_file)
        self.assertIn('EVERYSK_TEST_VAR_INHERIT: str', config_pyi_file)
        self.assertIn('DEBUG: bool', config_pyi_file)
        self.assertIn('EVERYSK_GOOGLE_CLOUD_LOCATION: str', config_pyi_file)
        self.assertIn('EVERYSK_GOOGLE_CLOUD_PROJECT: str', config_pyi_file)
        self.assertIn('REDIS_HOST: str', config_pyi_file)
        self.assertIn('REDIS_PORT: int', config_pyi_file)

    def test_none_default_value(self):
        # https://everysk.atlassian.net/browse/COD-3833
        self.assertIsNone(config.settings.EVERYSK_TEST_STR_DEFAULT_NONE)
        config.settings.EVERYSK_TEST_STR_DEFAULT_NONE = 'string'

    def test_undefined_default_value(self):
        # https://everysk.atlassian.net/browse/COD-3833
        self.assertEqual(config.settings.EVERYSK_TEST_STR_DEFAULT_UNDEFINED, Undefined)
        config.settings.EVERYSK_TEST_STR_DEFAULT_UNDEFINED = 'string'

    def test_re_pattern(self):
        self.assertEqual(config.settings.EVERYSK_TEST_RE_PATTERN, re.compile(r'[azAZ]'))

    def test_re_pattern_none(self):
        self.assertIsNone(config.settings.EVERYSK_TEST_RE_PATTERN_NONE)

    def test_re_pattern_undefined(self):
        self.assertEqual(config.settings.EVERYSK_TEST_RE_PATTERN_UNDEFINED, Undefined)

    def test_new_setting_from_environment(self):
        environ['EVERYSK_TEST_NEW_SETTING'] = '0'
        reload(config)
        self.assertEqual(config.settings.EVERYSK_TEST_NEW_SETTING, '0')

    def test_settings_context(self):
        self.assertTrue(config.settings.DEBUG)
        with config.settings as context_settings:
            self.assertTrue(context_settings)
            context_settings.DEBUG = False
            self.assertFalse(context_settings.DEBUG)
            self.assertTrue(config.settings.DEBUG)
        self.assertTrue(config.settings.DEBUG)

    def test_settings_value(self):
        self.assertEqual(config.settings.PROFILE, 'DEV')
        environ['PROFILE'] = 'LOCAL'
        reload(config)
        self.assertEqual(config.settings.PROFILE, 'LOCAL')
        config.settings.PROFILE = 'PROD'
        self.assertEqual(config.settings.PROFILE, 'PROD')
        self.assertEqual(config.Settings.PROFILE, 'DEV')
        self.assertEqual(getenv('PROFILE'), 'LOCAL')

    def test_setattr(self):
        settings = config.Settings(singleton=False)
        self.assertEqual(settings.PROFILE, 'DEV')
        settings.PROFILE = 'LOCAL'
        self.assertEqual(settings.PROFILE, 'LOCAL')

    def test_setattr_context(self):
        settings = config.Settings(singleton=False)
        context_settings = config.Settings(singleton=False)
        self.assertEqual(settings.PROFILE, 'DEV')
        self.assertEqual(context_settings.PROFILE, 'DEV')
        token = config.SettingsManager.settings.set(context_settings)
        settings.PROFILE = 'LOCAL'
        # Because the get uses the context we need to end it here
        config.SettingsManager.settings.reset(token)
        self.assertEqual(context_settings.PROFILE, 'LOCAL')
        self.assertEqual(settings.PROFILE, 'DEV')


def func(name: str, call: mock.MagicMock) -> None:
    # We wait to start at the same time inside de context
    sleep(0.1)
    call(f'{name}: {config.settings.PROFILE}')

class SettingsManagerTestCase(TestCase):

    def test_init_with_default_value(self):
        self.assertEqual(config.settings.PROFILE, 'DEV')
        with config.SettingsManager():
            self.assertEqual(config.settings.PROFILE, 'DEV')

        self.assertEqual(config.settings.PROFILE, 'DEV')

    def test_init_with_settings_as_param(self):
        settings = config.Settings(singleton=False)
        settings.PROFILE = 'LOCAL'
        self.assertEqual(config.settings.PROFILE, 'DEV')
        with config.SettingsManager(context_settings=settings):
            self.assertEqual(config.settings.PROFILE, 'LOCAL')

        self.assertEqual(config.settings.PROFILE, 'DEV')

    def test_context_settings(self):
        self.assertEqual(config.settings.PROFILE, 'DEV')
        with config.SettingsManager():
            self.assertEqual(config.settings.PROFILE, 'DEV')
            config.settings.PROFILE = 'LOCAL'
            self.assertEqual(config.settings.PROFILE, 'LOCAL')

        self.assertEqual(config.settings.PROFILE, 'DEV')

    def test_concurrency(self):
        call = mock.MagicMock()
        Thread(target=func, kwargs={'name': 'before_outer_thread', 'call': call}).start()
        with config.SettingsManager():
            config.settings.PROFILE = 'LOCAL'
            sleep(0.1)
            Thread(target=func, kwargs={'name': 'inner_thread', 'call': call}).start()
            # We wait to keep the context open
            sleep(0.3)

        Thread(target=func, kwargs={'name': 'after_outer_thread', 'call': call}).start()
        sleep(0.2)
        call.assert_has_calls([
            mock.call('before_outer_thread: DEV'),
            mock.call('inner_thread: LOCAL'),
            mock.call('after_outer_thread: DEV')
        ])



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api.http_client import HTTPClient, RequestsClient, new_default_http_client
from everysk.core.object import BaseDict
from everysk.core.unittests import TestCase, mock


###############################################################################
#   HTTP Client Test Case Implementation
###############################################################################
class HTTPClientTestCase(TestCase):

    def test_initialization_with_default_values(self):
        client = HTTPClient()
        self.assertEqual(client.timeout, 3600)
        self.assertEqual(client.verify_ssl_certs, True)
        self.assertEqual(client.allow_redirects, False)

    def test_initialization_with_custom_values(self):
        client = HTTPClient(1000, False, True)
        self.assertEqual(client.timeout, 1000)
        self.assertEqual(client.verify_ssl_certs, False)
        self.assertEqual(client.allow_redirects, True)

    def test_request(self):
        client = HTTPClient()
        with self.assertRaises(NotImplementedError):
            client.request("GET", 'http://example.com', {})

###############################################################################
#   Requests Client TestCase Implementation
###############################################################################
class RequestsClientTestCase(TestCase):

    def test_request_method(self):
        expected_status_code = 200
        expected_content = 'Test Content'

        client = RequestsClient()
        with mock.patch('requests.request') as mock_request:
            mock_response = mock.Mock()
            mock_response.content = expected_content
            mock_response.status_code = expected_status_code
            mock_request.return_value = mock_response

            status_code, content = client.request("GET", "http://example.com", {})

        self.assertEqual(status_code, expected_status_code)
        self.assertEqual(content, expected_content)
        mock_request.assert_called_once_with(method="GET", url="http://example.com", headers={}, params=None, timeout=3600, verify=True, allow_redirects=False)

    def test_request_method_with_base_dict(self):
        expected_status_code = 200
        expected_content = 'Test Content'

        client = RequestsClient()
        with mock.patch('requests.request') as mock_request:
            mock_response = mock.Mock()
            mock_response.content = expected_content
            mock_response.status_code = expected_status_code
            mock_request.return_value = mock_response

            status_code, content = client.request("POST", "http://example.com", {}, {}, BaseDict(key='value'))

        self.assertEqual(status_code, expected_status_code)
        self.assertEqual(content, expected_content)
        mock_request.assert_called_once_with(method="POST", url="http://example.com", headers={}, params={}, data='{"key": "value"}', timeout=3600, verify=True, allow_redirects=False)

###############################################################################
#   Standard HTTP Client TestCase Implementation
###############################################################################
class StandardHTTPClientTestCase(TestCase):
    @mock.patch('everysk.api.http_client.RequestsClient')
    def test_new_default_http_client(self, mock_requests_client):
        timeout = 3000
        verify_ssl_certs = False
        allow_redirects = True
        client = new_default_http_client(timeout=timeout, verify_ssl_certs=verify_ssl_certs, allow_redirects=allow_redirects)

        mock_requests_client.assert_called_once_with(timeout=timeout, verify_ssl_certs=verify_ssl_certs, allow_redirects=allow_redirects)
        self.assertEqual(client, mock_requests_client.return_value)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import Tuple

from everysk.config import settings
from everysk.sdk.base import BaseSDK


###############################################################################
#   Expression Class Implementation
###############################################################################
class Expression(BaseSDK):

    def get_tokens(self, expression: str, data_types: Tuple[str] = settings.ENGINES_EXPRESSION_DEFAULT_DATA_TYPES) -> frozenset:
        """
        Get the tokens of an expression. In other words, extract the
        elements used in the expression.

        Args:
            expression (str): The expression to get the tokens of.
            data_types (Tuple[str], optional): The data types to use for tokenization. Default is settings.ENGINES_EXPRESSION_DEFAULT_DATA_TYPES.

        Returns:
            frozenset: The tokens of the expression.

        Example:
            >>> expression = Expression()
            >>> expression.get_tokens('a + b')
            {'a', 'b'}
        """
        return self.get_response(self_obj=self ,params={'expression': expression, 'data_types': data_types})

    def solve(self, expression: str, user_args: dict):
        """
        Solve an expression with user arguments. Each key inside
        the `user_args` dictionary will be used as a value to solve
        the `expression` operation.

        Args:
            expression (str): The expression to solve.
            user_args (dict): The user arguments to use for solving.

        Returns:
            Any: The result of the expression.

        Example:
            >>> expression = Expression()
            >>> expression.solve('a + b', {'a': 1, 'b': 2})
            3
        """
        return self.get_response(self_obj=self, params={'expression': expression, 'user_args': user_args})



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.unittests import TestCase
from everysk.utils import bool_convert, search_key_on_dict


class SearchKeyTestCase(TestCase):

    def test_search_key(self):
        dct = {'key': 'value'}
        self.assertListEqual(list(search_key_on_dict('key', dct)), ['value'])

    def test_search_key_more_results(self):
        dct = {'key': 'value', 'key2': {'key': 'test'}, 'key1': 'not'}
        self.assertListEqual(list(search_key_on_dict('key', dct)), ['value', 'test'])

    def test_search_key_inside_another_dict(self):
        dct = {'key1': {'key2': {'key': 'test'}}}
        self.assertListEqual(list(search_key_on_dict('key', dct)), ['test'])

    def test_search_key_inside_list(self):
        dct = {'key3': [{'key1': {'key2': {'key': 'test'}}}]}
        self.assertListEqual(list(search_key_on_dict('key', dct)), ['test'])


class BoolConverterTestCase(TestCase):

    def test_clean_value_true(self):
        self.assertTrue(bool_convert(1))
        self.assertTrue(bool_convert(True))
        self.assertTrue(bool_convert('y'))
        self.assertTrue(bool_convert('yes'))
        self.assertTrue(bool_convert('on'))

    def test_clean_value_false(self):
        self.assertFalse(bool_convert(0))
        self.assertFalse(bool_convert(False))
        self.assertFalse(bool_convert('n'))
        self.assertFalse(bool_convert('no'))
        self.assertFalse(bool_convert('off'))



# -*- coding: utf_8 -*-
###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import Self, Any

from everysk.core.exceptions import HttpError, SDKError, FieldValueError
from everysk.core.http import HttpSDKPOSTConnection
from everysk.core.object import BaseDict
from everysk.sdk.entities.base_list import EntityList
from everysk.sdk.entities.portfolio.security import Security


###############################################################################
#   Securities Class Implementation
###############################################################################
class Securities(EntityList):

    _attr_type: Security = Security

    def _validate(self, value: dict) -> Security:
        """
        Validates a dictionary value and converts it into a `Security` object.

        Args:
            value (dict): A dictionary representing the security data.

        Returns:
            Security: A `Security` object created from the provided dictionary.

        Raises:
            FieldValueError: If the provided value is not a valid dictionary.
            TypeError: If the provided value is not of dictionary type.

        Example:
            >>> from my_module import Security
            >>> security_data = {'symbol': 'AAPL', 'name': 'Apple Inc.', 'price': 150.0}
            >>> validated_security = self._validate(security_data)
            >>> isinstance(validated_security, Security)
            True
        """
        if not isinstance(value, (dict, BaseDict)):
            raise FieldValueError(f"Security: The value must be a dict or BaseDict. {type(value)}") from TypeError
        return self._attr_type(**value)

    def validate(self) -> bool:
        """
        Validates the required fields of each security in the collection.

        This method iterates through the collection of securities and calls the
        `validate_required_fields` method for each security to ensure that all
        required fields are present and have valid values.

        Returns:
            bool: True if validation succeeds for all securities, otherwise False.

        Raises:
            FieldValueError: If the collection of securities is empty.

        Example:
            >>> from everysk.sdk.entities.portfolio.securities import Securities
            >>> security_data = [{'symbol': 'AAPL', 'name': 'Apple Inc.', 'price': 150.0},
            ...                  {'symbol': 'GOOGL', 'name': 'Alphabet Inc.', 'price': 2800.0}]
            >>> security_collection = Securities(security_data)
            >>> security_collection.validate()
            True
        """
        if len(self) == 0:
            raise FieldValueError('The quantity of securities cannot be zero.')

        for security in self:
            security.validate_required_fields()
        return True

    @staticmethod
    def from_lists(security_list: list[list[Any]]) -> Self:
        """
        Create a Securities collection from a list of lists representing securities.

        This static method takes a list of lists where each inner list represents a security.
        The first inner list (header) is used to extract attribute names, and the subsequent
        lists are used to create individual `Security` objects within a `Securities` collection.

        Args:
            security_list (list[list[Any]]): A list of lists where each inner list represents a security.

        Returns:
            Self: A `Securities` collection containing `Security` objects created from the provided data.

        Example:
            >>> from everysk.sdk.entities.portfolio.securities import Securities
            >>> security_data = [
            ...     ['symbol', 'name', 'price'],
            ...     ['AAPL', 'Apple Inc.', 150.0],
            ...     ['GOOGL', 'Alphabet Inc.', 2800.0]
            ... ]
            >>> securities_collection = Securities.from_lists(security_data)
            >>> len(securities_collection)
            2
        """
        header: list[str] = security_list.pop(0)
        return Securities([
            Securities._attr_type.from_list(sec_as_list, header)
            for sec_as_list in security_list
        ])

    @staticmethod
    def diff(securities_a: list[dict], securities_b: list[dict], accessor: str = 'comparable') -> dict:
        """
        Diff two securities list. The securities list are compared by the
        accessor provided. The accessor is used to compare the securities
        list. The accessor must be a key in the securities list.

        Args:
            securities_a (Securities): Securities list to compare.
            securities_b (Securities): Securities list to compare.
            accessor (str): Accessor to compare.

        Returns:
            dict: Diff result.
                Keys:
                    added_positions (Securities): Added positions.
                    removed_positions (Securities): Removed positions.
                    partial_positions (Securities): Partial positions.
                    equal_positions (Securities): Equal positions.
        """
        try:
            response: dict = HttpSDKPOSTConnection(
                cls_name='Securities',
                params={
                    'securities_a': securities_a,
                    'securities_b': securities_b,
                    'accessor': accessor
                }
            ).get_response()
        except HttpError as error:
            raise SDKError(error.msg) from error

        return response

    def consolidate(self, consolidation_keys) -> Self:
        """
        Consolidate securities list. If no consolidation keys are provided,
        the securities list is returned as is. Otherwise, the securities list
        is consolidated by the provided keys.

        Args:
            consolidation_keys (List[str]): List of keys to consolidate
            by (e.g. ['symbol', 'instrument_class']). If no keys are provided,
            the securities list is returned as is.

        Returns:
            Securities: Consolidated securities list.
        """
        try:
            response: dict = HttpSDKPOSTConnection(
                cls_name='Securities', params={'consolidation_keys': consolidation_keys}
            ).get_response()
        except HttpError as error:
            raise SDKError(error.msg) from error

        return type(self)(**response)

    def remove_errors(self) -> Self:
        """
        Create a new Securities collection with securities having a status other than 'ERROR'.

        This method filters the current Securities collection, creating a new collection that
        includes only securities with a status attribute other than 'ERROR'.

        Returns:
            Self: A new Securities collection containing securities without the 'ERROR' status.

        Example:
            >>> from everysk.sdk.entities.portfolio.securities import Securities
            >>> security_data = [{'symbol': 'AAPL', 'status': 'OK'},
            ...                  {'symbol': 'GOOGL', 'status': 'ERROR'},
            ...                  {'symbol': 'MSFT', 'status': 'OK'}]
            >>> securities_collection = Securities(security_data)
            >>> filtered_securities = securities_collection.remove_errors()
            >>> len(filtered_securities)
            2
        """
        return Securities([security for security in self if security.status != 'ERROR'])

    def to_lists(self, header: list[str] = None) -> list[list[Any]]:
        """
        Convert the Securities collection to a list of lists.

        This method converts each `Security` object in the collection to a list representation.

        Args:
            header (list[str], optional): A list of strings representing the header row. If not provided,
                the method will create a header based on the keys of the first security in the collection.

        Returns:
            list[list[Any]]: A list of lists representing the securities in the collection, including the header.

        Example:
            >>> from everysk.sdk.entities.portfolio.securities import Securities
            >>> security_data = [{'symbol': 'AAPL', 'name': 'Apple Inc.', 'price': 150.0},
            ...                  {'symbol': 'GOOGL', 'name': 'Alphabet Inc.', 'price': 2800.0}]
            >>> securities_collection = Securities(securities=security_data)
            >>> securities_list = securities_collection.to_lists()
            >>> len(securities_list)
            3  # Including the header row
        """
        if header is None:
            first_security: Security = self[0]
            extra_data: dict = first_security.get('extra_data') or {}
            header = list(first_security.keys()) + list(extra_data.keys())
            header = self._attr_type.sort_header(header)

        security_list: list[list[Any]] = [security.to_list(header=header) for security in self]
        security_list.insert(0, header)

        return security_list

    def to_list(self, add_class_path: bool = False, recursion: bool = False) -> list[dict]:
        """
        This method converts each `Security` object in the collection to a dictionary representation.

        Args:
            with_internals (bool, optional): Indicates whether to include internal attributes in the dictionaries.
                Defaults to True.

        Returns:
            list[dict]: A list of dictionaries representing the securities in the collection.

        Example:
            >>> from everysk.sdk.entities.portfolio.securities import Securities
            >>> security_data = [{'symbol': 'AAPL', 'name': 'Apple Inc.', 'price': 150.0},
            ...                  {'symbol': 'GOOGL', 'name': 'Alphabet Inc.', 'price': 2800.0}]
            >>> securities_collection = Securities(security_data)
            >>> securities_dict = securities_collection.to_list()
            >>> len(securities_dict)
            2
        """
        return [security.to_dict(add_class_path=add_class_path, recursion=recursion) for security in self if hasattr(security, 'to_dict')]



# -*- coding: utf_8 -*-
###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import Callable, Any, Self

from everysk.config import settings
from everysk.core.datetime import Date, DateTime
from everysk.core.fields import DateField, DictField, FloatField, ListField, StrField
from everysk.core.object import BaseDict, BaseDictConfig, CLASS_KEY
from everysk.core.exceptions import FieldValueError
from everysk.sdk.engines import cryptography
from everysk.sdk.entities.base import BaseSDK


###############################################################################
#   ExtraDataField Class Implementation
###############################################################################
class ExtraDataField(DictField):

    def clean_value(self, value: Any) -> Any:
        """
        This function is used to clean the value before it is assigned to the field.
        It can be used to ensure consistency in the data.
        """
        if isinstance(value, dict):
            value = BaseDict(**value)

        return super().clean_value(value)


###############################################################################
#   Security Class Implementation
###############################################################################
class Security(BaseDict, BaseSDK):
    """
    This class represents a security entity and its attributes.

    Attributes:
        status (StrField): The status of the security.
        id (StrField): The unique identifier of the security.
        symbol (StrField): The symbol of the security.
        quantity (FloatField): The quantity of the security.
        instrument_class (StrField): The class of the instrument.
        ticker (StrField): The ticker symbol of the security.
        label (StrField): The label of the security.
        name (StrField): The name of the security.
        isin (StrField): The ISIN (International Securities Identification Number) of the security.
        exchange (StrField): The exchange where the security is traded.
        currency (StrField): The currency of the security.
        fx_rate (FloatField): The foreign exchange rate of the security.
        market_price (FloatField): The market price of the security.
        premium (FloatField): The premium associated with the security.
        market_value (FloatField): The market value of the security.
        market_value_in_base (FloatField): The market value of the security in the base currency.
        instrument_type (StrField): The type of instrument.
        instrument_subtype (StrField): The subtype of the instrument.
        asset_class (StrField): The asset class of the security.
        asset_subclass (StrField): The asset subclass of the security.
        error_type (StrField): The type of error associated with the security.
        error_message (StrField): The error message associated with the security.
        maturity_date (DateField): The maturity date of the security.
        indexer (StrField): The indexer used for the security.
        percent_index (FloatField): The percentage index value.
        rate (FloatField): The rate value.
        coupon (FloatField): The coupon value.
        multiplier (FloatField): The multiplier value.
        underlying (StrField): The underlying asset of the security.
        series (StrField): The series of the security.
        option_type (StrField): The type of option (if applicable).
        strike (FloatField): The strike price (if applicable).
        issue_price (FloatField): The issue price of the security.
        issue_date (DateField): The issue date of the security.
        issuer (StrField): The issuer of the security.
        issuer_type (StrField): The type of issuer.
        cost_price (FloatField): The cost price of the security.
        unrealized_pl (FloatField): The unrealized profit or loss of the security.
        unrealized_pl_in_base (FloatField): The unrealized profit or loss of the security in the base currency.
        book (StrField): The book associated with the security.
        trader (StrField): The trader responsible for the security.
        trade_id (StrField): The trade identifier of the security.
        operation (StrField): The operation related to the security.
        accounting (StrField): The accounting information of the security.
        warranty (FloatField): The warranty associated with the security.
        return_date (DateField): The return date of the security.
        settlement (DateField): The settlement date of the security.
        look_through_reference (StrField): The look-through reference of the security.
        extra_data (DictField): Additional data associated with the security.
        hash (StrField): The hash value of the security.
        display (StrField): The display information of the security.
        comparable (StrField): The comparable information of the security.
        previous_quantity (FloatField): The previous quantity of the security (not sure about its use).

    Example:
        >>> security = Security()
        >>> security.symbol = 'AAPL'
        >>> security.quantity = 100
        >>> security.market_price = 150.0
        >>> security.currency = 'USD'
        >>> security.name = 'Apple Inc.'
        >>> security.status = 'OK'
        >>> security
        {
            'status': 'OK',
            'id': None,
            'symbol': 'AAPL',
            'quantity': 100.0,
            'instrument_class': None,
            'ticker': None,
            'label': None,
            'name': 'Apple Inc.',
            'isin': None,
            'exchange': None,
            'currency': 'USD',
            'fx_rate': None,
            'market_price': 150.0,
            'premium': None,
            'market_value': None,
            'market_value_in_base': None,
            'instrument_type': None,
            'instrument_subtype': None,
            'asset_class': None,
            'asset_subclass': None,
            'error_type': None,
            'error_message': None,
            'maturity_date': None,
            'indexer': None,
            'percent_index': None,
            'rate': None,
            'coupon': None,
            'multiplier': None,
            'underlying': None,
            'series': None,
            'option_type': None,
            'strike': None,
            'issue_price': None,
            'issue_date': None,
            'issuer': None,
            'issuer_type': None,
            'cost_price': None,
            'unrealized_pl': None,
            'unrealized_pl_in_base': None,
            'book': None,
            'trader': None,
            'trade_id': None,
            'operation': None,
            'accounting': None,
            'warranty': None,
            'return_date': None,
            'settlement': None,
            'look_through_reference': None,
            'extra_data': None,
            'hash': None,
            'display': None,
            'comparable': None,
            'previous_quantity': None
        }
    """
    class Config(BaseDictConfig):
        default_status = StrField(default=settings.SECURITY_STATUS_OK, readonly=True)
        valid_status = ListField(default=[settings.SECURITY_STATUS_OK, settings.SECURITY_STATUS_DELISTED], readonly=True)
        exclude_keys: frozenset[str] = frozenset(['_is_frozen', '_silent', '_errors', '_orderable_attributes'])

    _config: Config = None
    status = StrField()
    id = StrField(required_lazy=True, min_size=1, max_size=settings.SYMBOL_ID_MAX_LEN)

    symbol = StrField(required_lazy=True, max_size=settings.SYMBOL_MAX_LENGTH)
    quantity = FloatField(required_lazy=True, min_size=settings.QUANTITY_MIN_VALUE, max_size=settings.QUANTITY_MAX_VALUE)
    instrument_class = StrField()

    ticker = StrField()
    label = StrField(min_size=0, max_size=settings.LABEL_MAX_LENGTH)
    name = StrField()
    isin = StrField(min_size=0, max_size=settings.SYMBOL_MAX_LENGTH)
    exchange = StrField()
    currency = StrField()
    fx_rate = FloatField(min_size=settings.FX_RATE_MIN_VALUE, max_size=settings.FX_RATE_MAX_VALUE)
    market_price = FloatField(min_size=settings.PRICE_MIN_VALUE, max_size=settings.PRICE_MAX_VALUE)
    premium = FloatField()
    market_value = FloatField(min_size=settings.PRICE_MIN_VALUE, max_size=settings.PRICE_MAX_VALUE)
    market_value_in_base = FloatField()

    instrument_type = StrField()
    instrument_subtype = StrField()
    asset_class = StrField()
    asset_subclass = StrField()

    error_type = StrField(min_size=0, max_size=settings.PORTFOLIO_SECURITY_ERROR_TYPE_MAX_LEN)
    error_message = StrField(min_size=0, max_size=settings.PORTFOLIO_SECURITY_ERROR_MESSAGE_MAX_LEN)

    maturity_date = DateField(empty_is_none=True)
    indexer = StrField()
    percent_index = FloatField(min_size=settings.PRICE_MIN_VALUE, max_size=settings.PRICE_MAX_VALUE)
    rate = FloatField(min_size=settings.PRICE_MIN_VALUE, max_size=settings.PRICE_MAX_VALUE)
    coupon = FloatField(min_size=settings.PRICE_MIN_VALUE, max_size=settings.PRICE_MAX_VALUE)

    multiplier = FloatField(min_size=settings.MULTIPLIER_MIN_VALUE, max_size=settings.MULTIPLIER_MAX_VALUE)
    underlying = StrField()
    series = StrField()
    option_type = StrField()
    strike = FloatField(min_size=settings.PRICE_MIN_VALUE, max_size=settings.PRICE_MAX_VALUE)

    issue_price = FloatField(min_size=settings.PRICE_MIN_VALUE, max_size=settings.PRICE_MAX_VALUE)
    issue_date = DateField(empty_is_none=True)
    issuer = StrField()
    issuer_type = StrField()

    cost_price = FloatField(min_size=settings.PRICE_MIN_VALUE, max_size=settings.PRICE_MAX_VALUE)
    unrealized_pl = FloatField(min_size=settings.PRICE_MIN_VALUE, max_size=settings.PRICE_MAX_VALUE)
    unrealized_pl_in_base = FloatField(min_size=settings.PRICE_MIN_VALUE, max_size=settings.PRICE_MAX_VALUE)

    book = StrField()
    trader = StrField(min_size=0, max_size=settings.PORTFOLIO_SECURITY_TYPE_MAX_LEN)
    trade_id = StrField()

    operation = StrField()
    accounting = StrField()
    warranty = FloatField(min_size=settings.QUANTITY_MIN_VALUE, max_size=settings.QUANTITY_MAX_VALUE)

    return_date = DateField(empty_is_none=True)
    settlement = DateField(empty_is_none=True)

    look_through_reference = StrField()

    extra_data = ExtraDataField(default=None)

    hash = StrField()
    display = StrField()
    comparable = StrField()

    #NOT SURE ABOUT IT
    previous_quantity = FloatField(min_size=settings.QUANTITY_MIN_VALUE, max_size=settings.QUANTITY_MAX_VALUE)

    def __init__(self, **kwargs):
        kwargs.pop('id_', None)
        kwargs.pop('mic', None)
        kwargs.pop(CLASS_KEY, None)

        # This creates the Security with all default attributes
        super().__init__()

        # We check if the extra_data is not initialized
        _extra_data = kwargs.pop('extra_data', None) or BaseDict()
        if not isinstance(_extra_data, (BaseDict, dict)):
            raise FieldValueError(f'attribute extra_data must be a dictionary. {type(_extra_data)}')

        self.extra_data = BaseDict()
        if _extra_data is not None:
            for extra_key, extra_value in _extra_data.items():
                self._add_extra_data(extra_key, extra_value)

        for key, value in kwargs.items():
            # If the key already exists in the class we just store it
            if key in self:
                self[key] = value
            # otherwise we store it inside the extra_data key.
            else:
                self._add_extra_data(key, value)

        # If for some reason the extra_data is empty we convert it to None
        if not self.extra_data:
            self.extra_data = None

    def _add_extra_data(self, key: str, value: Any) -> None:
        """
        Add an extra data attribute to the security.

        This method adds an extra data attribute to the security object by storing it in the 'extra_data' dictionary.

        Args:
            key (str): The key for the extra data attribute.
            value (Any): The value of the extra data attribute.

        Example:
            >>> security = Security()
            >>> security._add_extra_data('industry', 'Technology')
            >>> security.extra_data
            {'industry': 'Technology'}
        """
        if value is None or isinstance(value, (int, str, bool, float, Date, DateTime)):
            self.extra_data[key] = value

    def _process_maturity_date(self, value: Date) -> str:
        """
        Converts a Date object to a string.

        Args:
            value (Date): The Date object to be converted.

        Returns:
            str: The converted Date, now represented as a string.

        Example:
            >>> _process_maturity_date(Date(2023, 9, 9))
            '20230909'
        """
        return Date.strftime_or_null(value)

    def _process_issue_date(self, value: Date) -> str:
        """
        Converts a Date object to a string.

        Args:
            value (Date): The Date object to be converted.

        Returns:
            str: The converted Date, now represented as a string.

        Example:
            >>> _process_issue_date(Date(2023, 9, 9))
            '20230909'
        """
        return Date.strftime_or_null(value)

    def _process_return_date(self, value: Date) -> str:
        """
        Converts a Date object to a string.

        Args:
            value (Date): The Date object to be converted.

        Returns:
            str: The converted Date, now represented as a string.

        Example:
            >>> _process_return_date(Date(2023, 9, 9))
            '20230909'
        """
        return Date.strftime_or_null(value)

    def _process_settlement(self, value: Date) -> str:
        """
        Converts a Date object to a string.

        Args:
            value (Date): The Date object to be converted.

        Returns:
            str: The converted Data, now represented as a string.

        Example:
            >>> _process_settlement(Date(2023, 9, 9))
            '20230909'
        """
        return Date.strftime_or_null(value)

    @staticmethod
    def _get_attr(security: dict, key: str, fallback_func: Callable | None = None) -> Any:
        """
        Get an attribute value from a security dictionary with optional fallback.

        This static method attempts to retrieve a value associated with a given key from a security dictionary.
        If the key is not found in the main dictionary, it checks if the key exists in the 'extra_data' dictionary
        within the security dictionary. If the key is still not found and a fallback function is provided, the
        fallback function is called to provide a default value.

        Args:
            security (dict): A dictionary representing a security and its attributes.
            key (str): The key for the attribute value to retrieve.
            fallback_func (Callable | None, optional): A fallback function to call if the key is not found in
                the main or 'extra_data' dictionaries. Defaults to None.

        Returns:
            Any: The attribute value if found, or the result of the fallback function (if provided), or None.

        Example:
            >>> from everysk.sdk.entities.portfolio.security import Security
            >>> security_data = {
            ...     'symbol': 'AAPL',
            ...     'name': 'Apple Inc.',
            ...     'extra_data': {'industry': 'Technology'},
            ... }
            >>> symbol = Security._get_attr(security_data, 'symbol')
            >>> industry = Security._get_attr(security_data, 'industry')
            >>> missing_attr = Security._get_attr(security_data, 'missing_attribute', lambda: 'Unknown')
            >>> symbol
            'AAPL'
            >>> industry
            'Technology'
            >>> missing_attr
            'Unknown'
        """
        if key in security:
            return security[key]

        extra_data: dict = security['extra_data'] or {}
        if key in extra_data:
            return extra_data[key]

        if fallback_func is not None:
            return fallback_func()

        return None

    @staticmethod
    def generate_security_id() -> str:
        """
        Generate a unique security ID.

        This static method generates a unique security ID by combining a prefix specified in
        'settings.SECURITY_ID_PREFIX' with a unique identifier of a specified length
        using the 'cryptography.generate_unique_id' function.

        Returns:
            str: A unique security ID.

        Example:
            >>> from everysk.sdk.entities.portfolio.security import Security
            >>> unique_id = Security.generate_security_id()
            >>> unique_id
            'SEC123456789'
        """
        return f'{settings.SECURITY_ID_PREFIX}{cryptography.generate_random_id(length=settings.SECURITY_ID_LENGTH)}'

    @staticmethod
    def sort_header(header: list[str]) -> list[str]:
        """
        Sort a header list based on a predefined order of properties.

        This static method takes a list of header strings and sorts them based on a predefined order
        specified in `settings.PORTFOLIO_PROPERTIES_ORDER`. Properties that are not found in the predefined
        order are sorted alphabetically at the end.

        Args:
            header (list[str]): A list of header strings to be sorted.

        Returns:
            list[str]: A sorted list of header strings.

        Example:
            >>> from everysk.sdk.entities.portfolio.security import Security
            >>> header = ['symbol', 'name', 'price', 'quantity']
            >>> sorted_header = Security.sort_header(header)
            >>> sorted_header
            ['quantity', 'name', 'price', 'symbol']
        """

        def sorting_key(element: str) -> tuple[int, str]:
            try:
                return (settings.PORTFOLIO_PROPERTIES_ORDER.index(element), '')
            except ValueError:
                return (len(settings.PORTFOLIO_PROPERTIES_ORDER), element)

        return sorted(header, key=sorting_key)

    def generate_consolidation_key(self, consolidation_keys: str) -> str:
        """
        Generate a consolidation key based on specified attributes.

        This method generates a consolidation key by extracting values from the security object
        based on the provided consolidation keys. It uses the `_get_attr` method to retrieve
        attribute values, and if a value is missing, it falls back to generating a security ID.

        Args:
            consolidation_keys (str): A string containing comma-separated attribute keys used for consolidation.

        Returns:
            str: The consolidation key generated from the specified attributes.

        Example:
            >>> from everysk.sdk.entities.portfolio.security import Security
            >>> security_data = {
            ...     'symbol': 'AAPL',
            ...     'name': 'Apple Inc.',
            ...     'industry': 'Technology',
            ...     'extra_data': {'country': 'USA'}
            ... }
            >>> security = Security(security_data)
            >>> consolidation_keys = 'symbol,name,industry,country'
            >>> consolidation_key = security.generate_consolidation_key(consolidation_keys)
            >>> consolidation_key
            'AAPL_Apple Inc._Technology_USA'
        """
        key_: list[Any] = [Security._get_attr(self, key, fallback_func=self.generate_security_id) for key in consolidation_keys]
        key: str = "_".join(str(v) for v in key_)

        return key

    @staticmethod
    def from_list(sec_as_list: list[Any], headers: list[str]) -> Self:
        """
        Create a Security object from a list of values and corresponding headers.

        This static method takes a list of values representing a security's attributes and a list
        of headers specifying the order of attributes. It constructs a Security object using the
        provided values and headers.

        Args:
            sec_as_list (list[Any]): A list of attribute values for the security.
            headers (list[str]): A list of headers specifying the order of attributes.

        Returns:
            Self: A new Security object initialized with the provided attribute values.

        Example:
            >>> from everysk.sdk.entities.portfolio.security import Security
            >>> sec_as_list = ['AAPL', 'Apple Inc.', 150.0, 100]
            >>> headers = ['symbol', 'name', 'market_price', 'quantity']
            >>> security = Security.from_list(sec_as_list, headers)
            >>> security
            {
                'symbol': 'AAPL',
                'name': 'Apple Inc.',
                'market_price': 150.0,
                'quantity': 100
            }
        """
        return Security(**dict(zip(headers, sec_as_list)))

    def validate_required_fields(self) -> bool:
        """
        Validate the required fields of the Security object.

        This method checks if the 'id' attribute is present in the Security object. If not, it generates
        a security ID using the 'generate_security_id' method and assigns it to the 'id' attribute.
        Then, it calls the superclass method to validate other required fields.

        Returns:
            bool: True if all required fields are valid, False otherwise.

        Example:
            >>> from everysk.sdk.entities.portfolio.security import Security
            >>> security = Security()
            >>> security.name = 'Apple Inc.'
            >>> security.quantity = 100
            >>> security.market_price = 150.0
            >>> security.validate_required_fields()
            True
        """
        if self.get('id', None) is None:
            self.id = self.generate_security_id()

        return super().validate_required_fields()

    def to_list(self, header: list[str] | None = None) -> list[Any]:
        """
        Convert the Security object to a list of attribute values.

        This method converts the Security object and its attributes to a list of values.
        The order of attributes in the list is determined by the 'header' parameter.
        If 'header' is not provided, it defaults to the sorted order specified by the
        'Security.sort_header' method.

        Args:
            header (list[str] | None, optional): A list of attribute keys specifying the order of attributes.
                Defaults to None.

        Returns:
            list[Any]: A list of attribute values for the Security object.

        Example:
            >>> from everysk.sdk.entities.portfolio.security import Security
            >>> security = Security()
            >>> security.symbol = 'AAPL'
            >>> security.quantity = 100
            >>> security.market_price = 150.0
            >>> security.name = 'Apple Inc.'
            >>> attribute_list = security.to_list()
            >>> attribute_list
            ['AAPL', 100, 150.0, 'Apple Inc.', # ... (other attributes)]
        """
        if header is None:
            header = Security.sort_header(self.keys())

        security: dict = self.to_dict(add_class_path=False)

        return [Security._get_attr(security, key) for key in header]

    def _process_extra_data(self, value: dict | BaseDict | None) -> dict:
        """
        Process the extra data attribute.
        """
        return value.to_dict() if hasattr(value, 'to_dict') else value



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.fields import StrField, IntField, RegexField


PRIVATE_SECURITY_SYMBOL_REGEX = RegexField(default=r'^PRIVATE:[A-Z0-9_]*$', readonly=True)
PRIVATE_SECURITY_SYMBOL_PREFIX = StrField(default='PRIVATE:', readonly=True)

PRIVATE_SECURITY_SYMBOL_MIN_SIZE = IntField(default=9, readonly=True) # len(PRIVATE_SECURITY_SYMBOL_PREFIX) + 1
PRIVATE_SECURITY_SYMBOL_MAX_SIZE = IntField(default=109, readonly=True) # 100 + len(PRIVATE_SECURITY_SYMBOL_PREFIX)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
import json
import zoneinfo

from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.datetime import DateTime, timezone
from everysk.core.exceptions import RequiredError, SDKValueError, FieldValueError
from everysk.core.fields import StrField
from everysk.core.http import HttpSDKPOSTConnection, httpx
from everysk.core.unittests import TestCase, mock

from everysk.sdk.entities.query import Query
from everysk.sdk.entities.base import BaseEntity
from everysk.sdk.entities.portfolio.base import Portfolio
from everysk.sdk.entities.portfolio.securities import Securities


###############################################################################
#   Base Entity TestCase Implementation
###############################################################################
class TestBaseEntity(TestCase):

    def setUp(self):
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()
        self.expected_response = {
            '_klass': Portfolio,
            'filters': [],
            'order': [],
            'projection': None,
            'limit': None,
            'offset': None,
            'page_size': None,
            'page_token': None,
            'distinct_on': []
        }
        self.sample_portfolio_data = {
            'id': 'port_1234567891011211234567890',
            'workspace': 'SampleWorkspace',
            'name': 'SamplePortfolio',
            'tags': ['tag1', 'tag2'],
            'link_uid': 'ABC',
            'description': 'Description',
            'nlv': 1000.0,
            'base_currency': 'USD',
            'date': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'securities': Securities([{'symbol': 'AAPL'}]),
            'version': '1.0',
            'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'level': 'HIGH',
            'outstanding_shares': 1000,
            'source_hash': 'XYZ',
            'status': 'OK',
            'portfolio_uid': 'UID123',
            'check_securities': False
        }
        self.portfolio = Portfolio(**self.sample_portfolio_data)

    def test_entity_to_query(self):
        portfolio = self.portfolio.copy()
        self.assertRaisesRegex(
            SDKValueError,
            "Can't filter by Name and Tags at the same time",
            portfolio.to_query
        )

        # test with name
        portfolio.update(tags=None, link_uid=None, date=None)
        _query = portfolio.to_query() # pylint: disable=protected-access
        self.assertDictEqual(_query,
        {
            '_klass': Portfolio,
            'filters': [
                ('workspace', '=', 'SampleWorkspace'),
                ('name', '<', 'sampleportfolip'),
                ('name', '>=', 'sampleportfolio')
            ],
            'order': [],
            'projection': None,
            'limit': None,
            'offset': None,
            'page_size': None,
            'page_token': None,
            'distinct_on': []
        })

        # test with tags
        portfolio.update(tags=['tag1', 'tag2'], name=None)
        _query = portfolio.to_query() # pylint: disable=protected-access
        self.assertListEqual(_query.filters, [
            ('workspace', '=', 'SampleWorkspace'),
            ('tags', '=', 'tag1'),
            ('tags', '=', 'tag2')
        ])

        # test with link_uid
        portfolio.update(tags=None, link_uid='ABC')
        _query = portfolio.to_query() # pylint: disable=protected-access
        self.assertListEqual(_query.filters, [
            ('workspace', '=', 'SampleWorkspace'),
            ('link_uid', '=', 'ABC')
        ])

    def test_check_query_raises_sdk_error_when_filtering_by_name_and_tags(self):
        with self.assertRaises(SDKValueError) as context:
            portfolio : Portfolio = self.portfolio.copy()
            portfolio.to_query(limit=1)

        self.assertEqual(str(context.exception), "Can't filter by Name and Tags at the same time")

    def test_check_query_raises_sdk_error_when_filtering_by_name_and_link(self):
        with self.assertRaises(SDKValueError) as context:
            portfolio: Portfolio = self.portfolio.copy()
            portfolio['tags'] = []
            portfolio.to_query(limit=1)

        self.assertEqual(str(context.exception), "Can't filter by Name and Link UID at the same time")

    def test_check_query_raises_sdk_error_when_date_in_query_order(self):
        with self.assertRaises(SDKValueError) as context:
            portfolio: Portfolio = self.portfolio.copy()
            portfolio['tags'] = []
            portfolio['link_uid'] = ''
            portfolio.to_query(order=['date'])

        self.assertEqual(str(context.exception), "Can't filter by Name and Date at the same time,  must order by updated_on")

    def test_get_id_prefix(self):
        with self.assertRaises(NotImplementedError):
            BaseEntity.get_id_prefix()

    def test_generate_id(self):
        with mock.patch.object(BaseEntity, 'get_id_prefix', return_value='P'), \
                mock.patch('everysk.sdk.entities.base.generate_random_id', return_value='12345'):
            self.assertEqual(BaseEntity().generate_id(), 'P12345')

    def test_validate_id(self):
        class FakeEntityClass(BaseEntity): # pylint: disable=abstract-method
            id = StrField(regex=settings.PORTFOLIO_ID_REGEX, max_size=settings.PORTFOLIO_ID_MAX_SIZE, required_lazy=True, empty_is_none=True)

        self.assertFalse(FakeEntityClass.validate_id('invalid_id'))
        self.assertTrue(FakeEntityClass.validate_id('port_puvGxrtIYc7OupBzfnrwFNVPg'))
        self.assertFalse(FakeEntityClass.validate_id(''))
        self.assertFalse(BaseEntity.validate_id(None))

    def test_validate(self):
        self.assertTrue(BaseEntity(id="my_id", workspace="my_workspace", name="my_name", created_on=DateTime.now(), updated_on=DateTime.now()).validate())

    def test_validate_transient(self):
        class FakeEntity(BaseEntity):
            @staticmethod
            def get_id_prefix():
                return 'fake_'

        entity = FakeEntity(workspace='my_workspace', name='my_name')
        entity['created_on'] = DateTime(2023, 9, 9, 9, 9, 9, 9)
        entity['updated_on'] = DateTime(2023, 9, 9, 9, 9, 9, 9)

        self.assertTrue(FakeEntity.validate_transient(entity))
        self.assertDictEqual(
            entity.__dict__,
            {'version': 'v1', 'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9, tzinfo=zoneinfo.ZoneInfo(key='UTC')), 'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9, tzinfo=zoneinfo.ZoneInfo(key='UTC')), 'workspace': 'my_workspace', 'id': None, 'name': 'my_name'}
        )

    def test_raises_validate(self):
        base_entity = BaseEntity()
        base_entity['created_on'] = None
        self.assertRaisesRegex(
            RequiredError,
            'The created_on attribute is required.',
            base_entity.validate
        )

    def test_get_query(self):
        """ Test _get_query method """
        expected_response = self.expected_response.copy()
        expected_response['projection'] = ['securities']
        result = Portfolio.query.set_projection('securities') # pylint: disable=protected-access
        self.assertDictEqual(result, expected_response)

    def test_query(self):
        """ Test query method """
        expected_response = self.expected_response.copy()
        expected_response['projection'] = ['securities']
        result = Portfolio.query.set_projection('securities')
        self.assertDictEqual(result, expected_response)

    def test_where(self):
        """ Test where method """
        expected_response = self.expected_response.copy()
        expected_response['filters'] = [('date', '=', DateTime(2023, 8, 10, 12, 0, tzinfo=timezone.utc))]
        result = Portfolio.query.where('date', '=', '2023-08-10')
        self.assertDictEqual(result, expected_response)

    def test_sort_by(self):
        """ Test sort_by method """
        expected_response = self.expected_response.copy()
        expected_response['order'] = ['date']
        result = Portfolio.query.sort_by('date')
        self.assertDictEqual(result, expected_response)

    def test_set_projection(self):
        """ Test set_projection method """
        expected_response = self.expected_response.copy()
        expected_response['projection'] = ['securities']
        result = Portfolio.query.set_projection('securities')
        self.assertDictEqual(result, expected_response)

    def test_set_limit(self):
        """ Test set_limit method """
        expected_response = self.expected_response.copy()
        expected_response['limit'] = 5
        result = Portfolio.query.set_limit(5)
        self.assertDictEqual(result, expected_response)

    def test_set_offset(self):
        """ Test set_offset method """
        expected_response = self.expected_response.copy()
        expected_response['offset'] = 10
        result = Portfolio.query.set_offset(10)
        self.assertDictEqual(result, expected_response)

    def test_set_page_size(self):
        """ Test set_page_size method """
        expected_response = self.expected_response.copy()
        expected_response['page_size'] = 15
        result = Portfolio.query.set_page_size(15)
        self.assertDictEqual(result, expected_response)

    def test_set_page_token(self):
        """ Test set_page_token method """
        expected_response = self.expected_response.copy()
        expected_response['page_token'] = 'my_test_page'
        result = Portfolio.query.set_page_token('my_test_page')
        self.assertDictEqual(result, expected_response)

    def test_load(self):
        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{"key": "value"}'
            mock_post.return_value.status_code = 200
            query = Portfolio.query.where('workspace', 'my_workspace')
            result = query.load(offset=5)

        expected_content = compress({'class_name': 'Query', 'method_name': 'load', 'self_obj': query.to_dict(add_class_path=True), 'params': {'offset': 5}}, protocol='gzip', serialize='json')
        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertIsInstance(result, Portfolio)

    def test_list(self):
        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '[{"key": "value"}]'
            mock_post.return_value.status_code = 200
            query = Portfolio.query.where('workspace', 'my_workspace')
            result = query.loads(limit=10, offset=5)

        expected_content = compress({'class_name': 'Query', 'method_name': 'loads', 'self_obj': query.to_dict(add_class_path=True), 'params': {'limit': 10, 'offset': 5}}, protocol='gzip', serialize='json')
        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 1)
        self.assertIsInstance(result[0], Portfolio)

    def test_page(self):
        query_class = Query(**self.expected_response)
        query_class['filters'] = [('workspace', '=', 'my_workspace')]

        expected_content = compress({'class_name': 'Query', 'method_name': 'page', 'self_obj': query_class.to_dict(add_class_path=True), 'params': {'page_size': 10, 'page_token': 'page_token'}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = json.dumps({
                "entities": [
                    {"key1": "value1", "key2": "value2", "key3": "value3"},
                ]
            })
            mock_post.return_value.status_code = 200
            result = Portfolio.query.where('workspace', 'my_workspace').page(page_size=10, page_token='page_token')

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertIsInstance(result['entities'][0], Portfolio)

    def test_pages(self):
        query = Query(**self.expected_response)
        query['filters'] = [('workspace', '=', 'my_workspace')]
        expected_content = compress({'class_name': 'Query', 'method_name': 'page', 'self_obj': query.to_dict(add_class_path=True), 'params': {'page_size': 10, 'page_token': {'__undefined__': None}}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = json.dumps({
                "entities": [
                    {"key1": "value1", "key2": "value2", "key3": "value3"},
                    {"key1": "value4", "key2": "value5", "key3": "value6"},
                    {"key1": "value7", "key2": "value8", "key3": "value9"}
                ]
            })
            mock_post.return_value.status_code = 200
            result = Portfolio.query.where('workspace', 'my_workspace').pages(page_size=10)
            next(result)

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)

    def test_script_query(self):
        expected_content = compress({'class_name': 'Script', 'method_name': 'inner_fetch', 'self_obj': {'_klass': 'Portfolio', '__class_path__': 'everysk.sdk.entities.script.Script'}, 'params': {'user_input': 'port_12456', 'variant': 'select', 'workspace': None}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = Portfolio.script.fetch(user_input='port_12456', variant='select', workspace=None)

        expected_content = compress({'class_name': 'Script', 'method_name': 'inner_fetch', 'self_obj': Portfolio.script.to_dict(add_class_path=True), 'params': {'user_input': 'port_12456', 'variant': 'select', 'workspace': None}}, protocol='gzip', serialize='json')
        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertIsInstance(result, Portfolio)

    def test_retrieve(self):
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'retrieve', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890', 'projection': None}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{"name": "Portfolio"}'
            mock_post.return_value.status_code = 200
            Portfolio.retrieve(entity_id=self.sample_portfolio_data['id'])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)

    def test_retrieve_with_projection(self):
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'retrieve', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890', 'projection': ['-securities']}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{"name": "Portfolio"}'
            mock_post.return_value.status_code = 200
            Portfolio.retrieve(entity_id=self.sample_portfolio_data['id'], projection='-securities')

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)

    def test_retrieve_many_with_invalid(self):
        self.assertRaisesRegex(
            FieldValueError,
            "The argument 'entity_id_list' most be a instance of 'list' and not <class 'str'>",
            Portfolio.retrieve_many,
            entity_id_list='port_1234567891011211234567890'
        )

    def test_inner_retrieve_many(self):
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'inner_retrieve_many', 'self_obj': None, 'params': {'entity_id_list': ['port_1234567891011211234567890'], 'projection': None}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '[{"name": "Portfolio"}, {"name": "Portfolio"}]'
            mock_post.return_value.status_code = 200
            Portfolio.retrieve_many(entity_id_list=[self.sample_portfolio_data['id']])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)

    def test_inner_retrieve_many_with_projection(self):
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'inner_retrieve_many', 'self_obj': None, 'params': {'entity_id_list': ['port_1234567891011211234567890'], 'projection': ['-securities', '-name']}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '[{"name": "Portfolio"}, {"name": "Portfolio"}]'
            mock_post.return_value.status_code = 200
            Portfolio.retrieve_many(entity_id_list=[self.sample_portfolio_data['id']], projection=['-securities', '-name'])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)

    def test_retrieve_with_none(self):
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'retrieve', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890', 'projection': None}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = 'null'
            mock_post.return_value.status_code = 200
            Portfolio.retrieve(entity_id=self.sample_portfolio_data['id'])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)

    def test_create(self):
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'create', 'self_obj': None, 'params': {'entity_dict': {'id': 'port_1234567891011211234567890', 'workspace': 'SampleWorkspace', 'name': 'SamplePortfolio', 'tags': ['tag1', 'tag2'], 'link_uid': 'ABC', 'description': 'Description', 'nlv': 1000.0, 'base_currency': 'USD', 'date': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'securities': [], 'version': '1.0', 'created_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'updated_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'level': 'HIGH', 'outstanding_shares': 1000, 'source_hash': 'XYZ', 'status': 'OK', 'portfolio_uid': 'UID123', 'check_securities': False}}}, protocol='gzip', serialize='json')
        self.sample_portfolio_data['securities'] = []

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{"creating": "portfolio"}'
            mock_post.return_value.status_code = 200
            Portfolio.create(entity_dict=self.sample_portfolio_data)

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)

    def test_modify(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'modify', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890', 'overwrites': {'base_currency': 'BRL'}}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            Portfolio.modify(entity_id=self.sample_portfolio_data['id'], overwrites={'base_currency': 'BRL'})

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)

    def test_modify_return_none(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'

        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'modify', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890', 'overwrites': {'base_currency': 'BRL'}}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = 'null'
            mock_post.return_value.status_code = 200
            Portfolio.modify(entity_id=self.sample_portfolio_data['id'], overwrites={'base_currency': 'BRL'})

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)

    def test_remove(self):
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'remove', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890'}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{"remove": "key"}'
            mock_post.return_value.status_code = 200
            result = Portfolio.remove(entity_id=self.sample_portfolio_data['id'])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertIsInstance(result, Portfolio)

    def test_remove_return_none(self):
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'remove', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890'}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = 'null'
            mock_post.return_value.status_code = 200
            result = Portfolio.remove(entity_id=self.sample_portfolio_data['id'])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertIsNone(result)

    def test_clone(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'clone', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890', 'overwrites': {'base_currency': 'BRL'}}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{"key": "value"}'
            mock_post.return_value.status_code = 200
            result = Portfolio.clone(entity_id=self.sample_portfolio_data['id'], overwrites={'base_currency': 'BRL'})

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertIsInstance(result, Portfolio)

    def test_clone_return_none(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'clone', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890', 'overwrites': {'base_currency': 'BRL'}}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = 'null'
            mock_post.return_value.status_code = 200
            result = Portfolio.clone(entity_id=self.sample_portfolio_data['id'], overwrites={'base_currency': 'BRL'})

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertIsNone(result)

    def test_create_many(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['base_currency'] = 'BRL'
        sample_portfolio_data['check_securities'] = True
        sample_portfolio_data['securities'] = []
        self.sample_portfolio_data['securities'] = []

        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'create_many', 'self_obj': None, 'params': {'entity_dict_list': [{'id': 'port_1234567891011211234567890', 'workspace': 'SampleWorkspace', 'name': 'SamplePortfolio', 'tags': ['tag1', 'tag2'], 'link_uid': 'ABC', 'description': 'Description', 'nlv': 1000.0, 'base_currency': 'USD', 'date': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'securities': [], 'version': '1.0', 'created_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'updated_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'level': 'HIGH', 'outstanding_shares': 1000, 'source_hash': 'XYZ', 'status': 'OK', 'portfolio_uid': 'UID123', 'check_securities': False}, {'id': 'port_1234567891011211234567890', 'workspace': 'SampleWorkspace', 'name': 'SamplePortfolio', 'tags': ['tag1', 'tag2'], 'link_uid': 'ABC', 'description': 'Description', 'nlv': 1000.0, 'base_currency': 'BRL', 'date': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'securities': [], 'version': '1.0', 'created_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'updated_on': {'__datetime__': '2023-09-09T09:09:09.000009+00:00'}, 'level': 'HIGH', 'outstanding_shares': 1000, 'source_hash': 'XYZ', 'status': 'OK', 'portfolio_uid': 'UID123', 'check_securities': True}]}}, protocol='gzip', serialize='json')
        expected_response = []

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = Portfolio.create_many(entity_dict_list=[self.sample_portfolio_data, sample_portfolio_data])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertEqual(result, expected_response)

    def test_modify_many(self):
        original_portfolio_data = self.sample_portfolio_data.copy()
        original_portfolio_data['name'] = 'modified_1'
        Portfolio(**original_portfolio_data)

        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['id'] = 'port_1234567891011211234567891'
        sample_portfolio_data['name'] = 'modified_2'
        Portfolio(**sample_portfolio_data)

        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'modify_many', 'self_obj': None, 'params': {'entity_id_list': ['port_1234567891011211234567890', 'port_1234567891011211234567891'], 'overwrites': [{'name': 'modified_1'}, {'name': 'modified_2'}]}}, protocol='gzip', serialize='json')
        expected_response = []

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = Portfolio.modify_many(entity_id_list=[self.sample_portfolio_data['id'], sample_portfolio_data['id']], overwrites=[{'name': 'modified_1'}, {'name': 'modified_2'}])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertEqual(result, expected_response)

    def test_remove_many(self):
        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['id'] = 'port_1234567891011211234567891'
        sample_portfolio_data['base_currency'] = 'BRL'

        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'remove_many', 'self_obj': None, 'params': {'entity_id_list': ['port_1234567891011211234567890', 'port_1234567891011211234567891']}}, protocol='gzip', serialize='json')
        expected_response = [self.sample_portfolio_data['id'], sample_portfolio_data['id']]

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = f'["{self.sample_portfolio_data["id"]}", "{sample_portfolio_data["id"]}"]'
            mock_post.return_value.status_code = 200
            result = Portfolio.remove_many(entity_id_list=[self.sample_portfolio_data['id'], sample_portfolio_data['id']])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertEqual(result, expected_response)

    def test_clone_many(self):
        original_portfolio_data = self.sample_portfolio_data.copy()
        original_portfolio_data['name'] = 'modified_1'
        original_portfolio_data['workspace'] = 'modified_1'
        original_portfolio = Portfolio(**original_portfolio_data)

        sample_portfolio_data = self.sample_portfolio_data.copy()
        sample_portfolio_data['id'] = 'port_1234567891011211234567891'
        sample_portfolio_data['name'] = 'modified_2'
        sample_portfolio_data['workspace'] = 'modified_2'
        Portfolio(**sample_portfolio_data)

        entity_id_list = [original_portfolio['id'], sample_portfolio_data['id']]

        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'clone_many', 'self_obj': None, 'params': {'entity_id_list': ['port_1234567891011211234567890', 'port_1234567891011211234567891'], 'overwrites': [{'name': 'modified_1'}, {'name': 'modified_1'}]}}, protocol='gzip', serialize='json')
        expected_response = []

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = Portfolio.clone_many(entity_id_list=entity_id_list, overwrites=[{'name': 'modified_1'}, {'name': 'modified_1'}])

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertEqual(result, expected_response)

    def test_save(self):
        self.sample_portfolio_data['securities'] = []
        portfolio = Portfolio(**self.sample_portfolio_data)
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'save', 'self_obj': portfolio.to_dict(add_class_path=True), 'params': {}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{"key": "value"}'
            mock_post.return_value.status_code = 200
            result = portfolio.save()

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertIsInstance(result, Portfolio)

    def test_delete(self):
        self.sample_portfolio_data['securities'] = []
        portfolio = Portfolio(**self.sample_portfolio_data)
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'delete', 'self_obj': portfolio.to_dict(add_class_path=True), 'params': {}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = portfolio.delete()

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertIsInstance(result, Portfolio)

    def test_delete_return_none(self):
        self.sample_portfolio_data['securities'] = []
        portfolio = Portfolio(**self.sample_portfolio_data)
        expected_content = compress({'class_name': 'Portfolio', 'method_name': 'delete', 'self_obj': portfolio.to_dict(add_class_path=True), 'params': {}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = 'null'
            mock_post.return_value.status_code = 200
            result = portfolio.delete()

        mock_post.assert_called_with(url=self.api_url, headers=self.headers, timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT), content=expected_content)
        self.assertIsNone(result)

    def test_obj_with_to_dict_as_attribute(self):
        port = Portfolio(securities=[{'symbol': 'AAPL', 'name': 'Apple Inc.', 'price': 150.0}])
        result = port.to_dict()

        self.assertIsInstance(result, dict)
        self.assertIsInstance(result['securities'], list)
        self.assertEqual(result['securities'][0]['symbol'], 'AAPL')
        self.assertNotEqual(result['securities'], [])

    def test_normalize_projection(self):
        self.assertListEqual(self.portfolio._normalize_projection(None), [])
        self.assertListEqual(self.portfolio._normalize_projection('securities'), ['securities'])
        self.assertListEqual(self.portfolio._normalize_projection('-securities'), ['-securities'])
        self.assertListEqual(self.portfolio._normalize_projection(['-securities', '-name']), ['-securities', '-name'])



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.fields import StrField, IntField, RegexField


REPORT_ID_REGEX = RegexField(default=r'^repo_[a-zA-Z0-9]', readonly=True)
REPORT_ID_PREFIX = StrField(default='repo_', readonly=True)

REPORT_ID_MAX_SIZE = IntField(default=30, readonly=True)
REPORT_LEVEL_MAX_LENGTH = IntField(default=32, readonly=True)

REPORT_AUTHORIZATION_PUBLIC = StrField(default='public', readonly=True)
REPORT_AUTHORIZATION_PRIVATE = StrField(default='private', readonly=True)
REPORT_AUTHORIZATION_REGEX = RegexField(default=r'public|private', readonly=True)

REPORT_URL_PATH = StrField(default='/report', readonly=True)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=protected-access
import json
import logging
import traceback
import sys
from logging import Logger as PythonLogger, DEBUG, StreamHandler
from time import sleep
from everysk.config import settings
from everysk.core import log as log_module
from everysk.core.unittests import TestCase, mock


@mock.patch.dict('os.environ', {'LOGGING_JSON': '1'})
class LoggerJsonTestCase(TestCase):
    # The first TestCase so we don't need to change the line number in the tests
    # every time we add a new test or lines

    def test_info_as_json(self):
        log = log_module.Logger(name='everysk-log-json-test')
        with mock.patch.object(log._log.handler, 'stream') as stream:
            log.info('info')
            stream.write.assert_called_once_with(
                '{"message": "info", '
                '"severity": "INFO", '
                '"logName": "everysk-log-json-test", '
                '"labels": {}, '
                '"traceback": "", '
                '"http": {"headers": {}, "payload": {}, "response": {}}, '
                '"logging.googleapis.com/trace": "", '
                '"logging.googleapis.com/spanId": "", '
                '"logging.googleapis.com/trace_sampled": false, '
                '"logging.googleapis.com/sourceLocation":'
                ' {"file": "/var/app/src/everysk/core/_tests/log.py", "line": 30, "function": "test_info_as_json"}}\n'
            )

    def test_info_as_json_with_labels(self):
        log = log_module.Logger(name='everysk-log-json-test')
        with mock.patch.object(log._log.handler, 'stream') as stream:
            log.info('info with labels', extra={'labels': {'key': 'value'}})
            stream.write.assert_called_once_with(
                '{"message": "info with labels", '
                '"severity": "INFO", '
                '"logName": "everysk-log-json-test", '
                '"labels": {"key": "value"}, '
                '"traceback": "", '
                '"http": {"headers": {}, "payload": {}, "response": {}}, '
                '"logging.googleapis.com/trace": "", '
                '"logging.googleapis.com/spanId": "", '
                '"logging.googleapis.com/trace_sampled": false, '
                '"logging.googleapis.com/sourceLocation":'
                ' {"file": "/var/app/src/everysk/core/_tests/log.py", "line": 48, "function": "test_info_as_json_with_labels"}}\n'
            )

    def test_info_as_json_with_labels_as_bytes(self):
        log = log_module.Logger(name='everysk-log-json-test')
        with mock.patch.object(log._log.handler, 'stream') as stream:
            log.info('info with labels as bytes', extra={'labels': b'text'})
            stream.write.assert_called_once_with(
                '{"message": "info with labels as bytes", '
                '"severity": "INFO", '
                '"logName": "everysk-log-json-test", '
                '"labels": "text", '
                '"traceback": "", '
                '"http": {"headers": {}, "payload": {}, "response": {}}, '
                '"logging.googleapis.com/trace": "", '
                '"logging.googleapis.com/spanId": "", '
                '"logging.googleapis.com/trace_sampled": false, '
                '"logging.googleapis.com/sourceLocation":'
                ' {"file": "/var/app/src/everysk/core/_tests/log.py", "line": 66, "function": "test_info_as_json_with_labels_as_bytes"}}\n'
            )

    def test_info_as_json_with_labels_with_bytes(self):
        log = log_module.Logger(name='everysk-log-json-test')
        with mock.patch.object(log._log.handler, 'stream') as stream:
            log.info('info with labels with bytes', extra={'labels': {'key': b'text'}})
            stream.write.assert_called_once_with(
                '{"message": "info with labels with bytes", '
                '"severity": "INFO", '
                '"logName": "everysk-log-json-test", '
                '"labels": {"key": "text"}, '
                '"traceback": "", '
                '"http": {"headers": {}, "payload": {}, "response": {}}, '
                '"logging.googleapis.com/trace": "", '
                '"logging.googleapis.com/spanId": "", '
                '"logging.googleapis.com/trace_sampled": false, '
                '"logging.googleapis.com/sourceLocation":'
                ' {"file": "/var/app/src/everysk/core/_tests/log.py", "line": 84, "function": "test_info_as_json_with_labels_with_bytes"}}\n'
            )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_info_as_json_with_http_headers(self):
        log = log_module.Logger(name='everysk-log-json-test')
        with mock.patch.object(log._log.handler, 'stream') as stream:
            log.info('info with http headers', extra={'http_headers': {'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'}})
            stream.write.assert_called_once_with(
                '{"message": "info with http headers", '
                '"severity": "INFO", '
                '"logName": "everysk-log-json-test", '
                '"labels": {}, '
                '"traceback": "", '
                '"http": {"headers": {"traceparent": "00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01"}, "payload": {}, "response": {}}, '
                '"logging.googleapis.com/trace": "projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df", '
                '"logging.googleapis.com/spanId": "1c6c592f9e46e3fb", '
                '"logging.googleapis.com/trace_sampled": true, '
                '"logging.googleapis.com/sourceLocation":'
                ' {"file": "/var/app/src/everysk/core/_tests/log.py", "line": 103, "function": "test_info_as_json_with_http_headers"}}\n'
            )

    def test_error_as_json(self):
        log = log_module.Logger(name='everysk-log-json-test')
        with mock.patch.object(log._log.handler, 'stream') as stream:
            trace = ''
            try:
                raise ValueError('test')
            except ValueError:
                trace = traceback.format_exc()
                log.error('error')
            stream.write.assert_called_once_with(
                '{"message": "error", '
                '"severity": "ERROR", '
                '"logName": "everysk-log-json-test", '
                '"labels": {}, '
                f'"traceback": {json.dumps(trace)}, '
                '"http": {"headers": {}, "payload": {}, "response": {}}, '
                '"logging.googleapis.com/trace": "", '
                '"logging.googleapis.com/spanId": "", '
                '"logging.googleapis.com/trace_sampled": false, '
                '"logging.googleapis.com/sourceLocation":'
                ' {"file": "/var/app/src/everysk/core/_tests/log.py", "line": 126, "function": "test_error_as_json"}}\n'
            )


class LoggerTraceTestCase(TestCase):

    def setUp(self) -> None:
        self.trace_parts = log_module.DEFAULT_TRACE_PARTS.copy()

    def test_traceparent(self):
        traceparent = '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'
        log_module._parse_traceparent(traceparent, self.trace_parts)
        self.assertDictEqual(
            self.trace_parts,
            {
                'span_id': '1c6c592f9e46e3fb',
                'trace_id': '4bfa9e049143840bef864a7859f2e5df',
                'trace_sample': True,
                'version': '00',
            }
        )

    def test_parse_traceparent_invalid(self):
        traceparent = 'invalid'
        log_module._parse_traceparent(traceparent, self.trace_parts)
        self.assertDictEqual(
            self.trace_parts,
            {
                'span_id': '',
                'trace_id': '',
                'trace_sample': False,
                'version': '',
            }
        )

    def test_parse_x_cloud_trace_context(self):
        trace_context = '4bfa9e049143840bef864a7859f2e5df/2048109991600514043;o=1'
        log_module._parse_x_cloud_trace_context(trace_context, self.trace_parts)
        self.assertDictEqual(
            self.trace_parts,
            {
                'span_id': '2048109991600514043',
                'trace_id': '4bfa9e049143840bef864a7859f2e5df',
                'trace_sample': True,
                'version': '',
            }
        )

    def test_parse_x_cloud_trace_context_invalid(self):
        trace_context = 'invalid'
        log_module._parse_x_cloud_trace_context(trace_context, self.trace_parts)
        self.assertDictEqual(
            self.trace_parts,
            {'span_id': '', 'trace_id': '', 'trace_sample': False, 'version': ''}
        )

    def test_get_trace_data_no_headers(self):
        self.assertDictEqual(
            log_module._get_trace_data(headers={}),
            {'span_id': '', 'trace_id': '', 'trace_sample': False, 'version': ''}
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_trace_data_traceparent(self):
        self.assertDictEqual(
            log_module._get_trace_data(headers={'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'}),
            {
                'span_id': '1c6c592f9e46e3fb',
                'trace_id': 'projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df',
                'trace_sample': True,
                'version': '00',
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_trace_data_x_cloud_trace_context(self):
        self.assertDictEqual(
            log_module._get_trace_data(headers={'x-cloud-trace-context': '4bfa9e049143840bef864a7859f2e5df/2048109991600514043;o=1'}),
            {
                'span_id': '2048109991600514043',
                'trace_id': 'projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df',
                'trace_sample': True,
                'version': '',
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_trace_data_both(self):
        self.assertDictEqual(
            log_module._get_trace_data(headers={
                'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01',
                'x-cloud-trace-context': '4bfa9e049143840bef864a7859f2e5df/2048109991600514043;o=1'
            }),
            {
                'span_id': '1c6c592f9e46e3fb',
                'trace_id': 'projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df',
                'trace_sample': True,
                'version': '00',
            }
        )

    def test_get_traceback(self):
        try:
            raise ValueError('test')
        except ValueError:
            self.assertEqual(log_module._get_traceback(), traceback.format_exc())

    def test_get_traceback_no_traceback(self):
        self.assertEqual(log_module._get_traceback(), '')


class LoggerManagerTestCase(TestCase):

    def test_http_headers(self):
        with log_module.LoggerManager(http_headers={'X-Request-ID': '123'}):
            self.assertEqual(log_module.LoggerManager._extra.get(), {'http_headers': {'X-Request-ID': '123'}})

    def test_http_payload(self):
        with log_module.LoggerManager(http_payload={'data': '123'}):
            self.assertEqual(log_module.LoggerManager._extra.get(), {'http_payload': {'data': '123'}})

    def test_labels(self):
        with log_module.LoggerManager(labels={'data': '123'}):
            self.assertEqual(log_module.LoggerManager._extra.get(), {'labels': {'data': '123'}})

    def test_stacklevel(self):
        with log_module.LoggerManager(stacklevel=1):
            self.assertEqual(log_module.LoggerManager._extra.get(), {'stacklevel': 1})

    def test_traceback(self):
        trace = ''
        try:
            raise ValueError('test')
        except ValueError:
            trace = traceback.format_exc()

        with log_module.LoggerManager(traceback=trace):
            self.assertEqual(log_module.LoggerManager._extra.get(), {'traceback': trace})

    def test_empty(self):
        with log_module.LoggerManager():
            self.assertEqual(log_module.LoggerManager._extra.get(), {})

    def test_chain_managers(self):
        with log_module.LoggerManager(http_headers={'X-Request-ID': '123'}):
            with log_module.LoggerManager(http_payload={'data': '123'}):
                with log_module.LoggerManager(labels={'data': '123'}):

                    self.assertEqual(
                        log_module.LoggerManager._extra.get(),
                        {'http_headers': {'X-Request-ID': '123'}, 'http_payload': {'data': '123'}, 'labels': {'data': '123'}}
                    )
                self.assertEqual(
                    log_module.LoggerManager._extra.get(),
                    {'http_headers': {'X-Request-ID': '123'}, 'http_payload': {'data': '123'}}
                )
            self.assertEqual(
                log_module.LoggerManager._extra.get(),
                {'http_headers': {'X-Request-ID': '123'}}
            )

        self.assertEqual(log_module.LoggerManager._extra.get(), {})

    def test_reset(self):
        with log_module.LoggerManager(http_headers={'X-Request-ID': '123'}):
            log_module.LoggerManager.reset()
            self.assertEqual(log_module.LoggerManager._extra.get(), {})


class LoggerTestCase(TestCase):

    def test_logger_return(self):
        log = log_module.Logger('everysk-log-stdout-test')
        self.assertIsInstance(log._log, PythonLogger)

    def test_log_level(self):
        log = log_module.Logger(name='everysk-log-stdout-test')
        self.assertEqual(log._log.level, DEBUG)

    def test_log_propagate(self):
        log = log_module.Logger('everysk-log-stdout-test')
        self.assertFalse(log._log.propagate)

    def test_stdout_handler(self):
        log = log_module.Logger('everysk-log-stdout-test')
        handler = log._log.handlers[0]
        self.assertEqual(type(handler), StreamHandler)

    def test_handler_level(self):
        log = log_module.Logger(name='everysk-log-stdout-test')
        handler = log._log.handlers[0]
        self.assertEqual(handler.level, DEBUG)

    @mock.patch.dict('os.environ', {'LOGGING_JSON': 'false'})
    def test_handler_format(self):
        log = log_module.Logger(name='everysk-log-stdout-test')
        handler = log._log.handlers[0]
        self.assertEqual(
            handler.formatter._fmt,
            '%(asctime)s - %(levelname)s - %(labels)s - %(message)s'
        )

    def test_default_root_name(self):
        # https://everysk.atlassian.net/browse/COD-3212
        with self.assertRaisesRegex(ValueError, 'The name of the log could not be "root".'):
            log_module.Logger(name='root')


class LoggerExtraDataTestCase(TestCase):

    def setUp(self) -> None:
        self.log = log_module.Logger('everysk-log-extra-test')

    def test_http_headers(self):
        self.assertDictEqual(
            self.log._get_extra_data({'http_headers': {'X-Request-ID': '123'}}),
            {'labels': {}}
        )

    def test_http_headers_traceparent(self):
        self.assertDictEqual(
            self.log._get_extra_data({'http_headers': {'traceparent': '123'}}),
            {'http_headers': {'traceparent': '123'}, 'labels': {}}
        )

    def test_http_headers_x_cloud_trace_context(self):
        self.assertDictEqual(
            self.log._get_extra_data({'http_headers': {'x-cloud-trace-context': '123'}}),
            {'http_headers': {'x-cloud-trace-context': '123'}, 'labels': {}}
        )

    def test_http_payload(self):
        self.assertDictEqual(
            self.log._get_extra_data({'http_payload': {'data': '123'}}),
            {'http_payload': {'data': '123'}, 'labels': {}}
        )

    def test_another_key(self):
        self.assertDictEqual(
            self.log._get_extra_data({'another_key': '123'}),
            {'another_key': '123', 'labels': {}}
        )

    def test_http_headers_as_param(self):
        func = mock.MagicMock(return_value={'traceparent': 'flask'})
        self.log._http_functions = {'flask': {'http_headers': func}}
        with log_module.LoggerManager(http_headers={'traceparent': 'manager'}):
            self.assertDictEqual(
                self.log._get_extra_data({'http_headers': {'traceparent': 'param'}}, logging.ERROR),
                {'http_headers': {'traceparent': 'param'}, 'labels': {}}
            )
        func.assert_not_called()

    def test_http_headers_as_manager(self):
        func = mock.MagicMock(return_value={'traceparent': 'flask'})
        self.log._http_functions = {'flask': {'http_headers': func}}
        with log_module.LoggerManager(http_headers={'traceparent': 'manager'}):
            self.assertDictEqual(
                self.log._get_extra_data({}, logging.ERROR),
                {'http_headers': {'traceparent': 'manager'}, 'labels': {}}
            )
        func.assert_not_called()

    def test_http_payload_as_param(self):
        func = mock.MagicMock(return_value={'data': 'flask'})
        self.log._http_functions = {'flask': {'http_payload': func}}
        with log_module.LoggerManager(http_payload={'data': 'manager'}):
            self.assertDictEqual(
                self.log._get_extra_data({'http_payload': {'data': 'param'}}, logging.ERROR),
                {'http_payload': {'data': 'param'}, 'labels': {}}
            )
        func.assert_not_called()

    def test_http_payload_as_manager(self):
        func = mock.MagicMock(return_value={'data': 'flask'})
        self.log._http_functions = {'flask': {'http_payload': func}}
        with log_module.LoggerManager(http_payload={'data': 'manager'}):
            self.assertDictEqual(
                self.log._get_extra_data({}, logging.ERROR),
                {'http_payload': {'data': 'manager'}, 'labels': {}}
            )
        func.assert_not_called()

    def test_http_response_as_param(self):
        func = mock.MagicMock(return_value={'data': 'flask'})
        self.log._http_functions = {'flask': {'http_response': func}}
        with log_module.LoggerManager(http_response={'data': 'manager'}):
            self.assertDictEqual(
                self.log._get_extra_data({'http_response': {'data': 'param'}}, logging.ERROR),
                {'http_response': {'data': 'param'}, 'labels': {}}
            )
        func.assert_not_called()

    def test_http_response_as_manager(self):
        func = mock.MagicMock(return_value={'data': 'flask'})
        self.log._http_functions = {'flask': {'http_response': func}}
        with log_module.LoggerManager(http_response={'data': 'manager'}):
            self.assertDictEqual(
                self.log._get_extra_data({}, logging.ERROR),
                {'http_response': {'data': 'manager'}, 'labels': {}}
            )
        func.assert_not_called()

    def test_traceback_with_manager(self):
        trace = ''
        try:
            raise ValueError('test')
        except ValueError:
            trace = traceback.format_exc()

        try:
            a = 1 / 0 # pylint: disable=unused-variable
        except ZeroDivisionError:
            with log_module.LoggerManager(traceback=trace):
                self.assertDictEqual(
                    self.log._get_extra_data({}, logging.ERROR),
                    {'traceback': trace, 'labels': {}}
                )

    def test_traceback(self):
        try:
            raise ValueError('test')
        except ValueError:
            self.assertDictEqual(
                self.log._get_extra_data({}, logging.ERROR),
                {'traceback': traceback.format_exc(), 'labels': {}}
            )

    def test_traceback_as_param(self):
        try:
            raise ValueError('test')
        except ValueError:
            self.assertDictEqual(
                self.log._get_extra_data({'traceback': 'Teste'}, logging.ERROR),
                {'traceback': 'Teste', 'labels': {}}
            )

    def test_labels_with_manager(self):
        with log_module.LoggerManager(labels={'data': '123'}):
            self.assertDictEqual(
                self.log._get_extra_data({}, logging.ERROR),
                {'labels': {'data': '123'}}
            )

    def test_labels_method(self):
        self.assertDictEqual(
            self.log._get_extra_data({'labels': {'data': '123'}}, logging.ERROR),
            {'labels': {'data': '123'}}
        )


@mock.patch.object(log_module.Logger, '_log')
class LoggerMethodsTestCase(TestCase):

    def test_critical(self, log_mock: mock.MagicMock):
        log = log_module.Logger('everysk-log-methods-test')
        log.critical('critical')
        log_mock.log.assert_called_once_with(logging.CRITICAL, 'critical', extra={'labels': {}}, stacklevel=3)

    def test_debug(self, log_mock: mock.MagicMock):
        log = log_module.Logger('everysk-log-methods-test')
        log.debug('debug')
        log_mock.log.assert_called_once_with(logging.DEBUG, 'debug', extra={'labels': {}}, stacklevel=3)

    def test_error(self, log_mock: mock.MagicMock):
        log = log_module.Logger('everysk-log-methods-test')
        log.error('error')
        log_mock.log.assert_called_once_with(logging.ERROR, 'error', extra={'labels': {}}, stacklevel=3)

    def test_exception(self, log_mock: mock.MagicMock):
        log = log_module.Logger('everysk-log-methods-test')
        log.exception('error')
        log_mock.log.assert_called_once_with(logging.ERROR, 'error', extra={'labels': {}}, stacklevel=3)

    def test_info(self, log_mock: mock.MagicMock):
        log = log_module.Logger('everysk-log-methods-test')
        log.info('info')
        log_mock.log.assert_called_once_with(logging.INFO, 'info', extra={'labels': {}}, stacklevel=3)

    def test_warning(self, log_mock: mock.MagicMock):
        log = log_module.Logger('everysk-log-methods-test')
        log.warning('warning')
        log_mock.log.assert_called_once_with(logging.WARNING, 'warning', extra={'labels': {}}, stacklevel=3)

    def test_deprecated(self, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-log-deprecated')
        log.deprecated('Teste')
        log_mock.log.assert_called_once_with(logging.WARNING, 'DeprecationWarning: Teste', extra={'labels': {}}, stacklevel=3)

    def test_deprecated_once(self, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-log-deprecated')
        log.deprecated('Teste 1', show_once=True)
        log.deprecated('Teste 1', show_once=True)
        log_mock.log.assert_called_once_with(logging.WARNING, 'DeprecationWarning: Teste 1', extra={'labels': {}}, stacklevel=3)

    def test_deprecated_twice(self, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-log-deprecated')
        log.deprecated('Teste 2', show_once=False)
        log.deprecated('Teste 2', show_once=False)
        log_mock.log.assert_has_calls([
            mock.call(logging.WARNING, 'DeprecationWarning: Teste 2', extra={'labels': {}}, stacklevel=3),
            mock.call(logging.WARNING, 'DeprecationWarning: Teste 2', extra={'labels': {}}, stacklevel=3),
        ])


@mock.patch.object(log_module.Logger, '_log')
@mock.patch('everysk.core.slack.Slack')
class LoggerSlackTestCase(TestCase):

    def setUp(self) -> None:
        self.old_profile = settings.PROFILE
        self.old_slack_url = settings.SLACK_URL
        settings.PROFILE = 'PROD'
        settings.SLACK_URL = 'http://localhost'
        # This removes the module from the cache
        sys.modules.pop('unittest')

    def tearDown(self) -> None:
        settings.PROFILE = self.old_profile
        settings.SLACK_URL = self.old_slack_url
        # Put back the module
        import unittest # pylint: disable=unused-import, import-outside-toplevel

    def test_danger(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Error', message='Error message', color='danger')
        log_mock.log.assert_called_once_with(logging.ERROR, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=3)
        slack.assert_called_once_with(title='Error', message='Error message', color='danger', url=settings.SLACK_URL)

    def test_warning(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Warning', message='Warning message', color='warning')
        log_mock.log.assert_called_once_with(logging.WARNING, 'Slack message: Warning -> Warning message', extra={'labels': {}}, stacklevel=3)
        slack.assert_called_once_with(title='Warning', message='Warning message', color='warning', url=settings.SLACK_URL)

    def test_success(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Success', message='Success message', color='success')
        log_mock.log.assert_called_once_with(logging.INFO, 'Slack message: Success -> Success message', extra={'labels': {}}, stacklevel=3)
        slack.assert_called_once_with(title='Success', message='Success message', color='success', url=settings.SLACK_URL)

    def test_not_url(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        settings.SLACK_URL = None
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Error', message='Error message', color='danger')
        log_mock.log.assert_called_once_with(logging.ERROR, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=3)
        slack.assert_not_called()

    def test_not_prod(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        settings.PROFILE = 'DEV'
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Error', message='Error message', color='danger')
        log_mock.log.assert_called_once_with(logging.ERROR, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=3)
        slack.assert_not_called()

    def test_in_tests(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        import unittest # pylint: disable=unused-import, import-outside-toplevel
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Error', message='Error message', color='danger')
        log_mock.log.assert_called_once_with(logging.ERROR, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=3)
        slack.assert_not_called()

    def test_method_url(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Error', message='Error message', color='danger', url='http://google.com')
        log_mock.log.assert_called_once_with(logging.ERROR, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=3)
        slack.assert_called_once_with(title='Error', message='Error message', color='danger', url='http://google.com')

    def test_slack_time_check(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Error', message='Error message', color='danger', url='http://google.com')
        log.slack(title='Error', message='Error message', color='danger', url='http://google.com')
        slack.assert_called_once_with(title='Error', message='Error message', color='danger', url='http://google.com')
        log_mock.log.assert_has_calls([
            mock.call(40, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=3),
            mock.call(40, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=3)
        ])

    def test_slack_time_check_3s(self, slack: mock.MagicMock, log_mock: mock.MagicMock):
        log = log_module.Logger(name='everysk-lib-test-slack-message')
        log.slack(title='Error', message='Error message', color='danger', url='http://google.com')
        log.slack(title='Error', message='Error message', color='danger', url='http://google.com')
        sleep(3)
        log.slack(title='Error', message='Error message', color='danger', url='http://google.com')
        slack.assert_has_calls([
            mock.call(title='Error', message='Error message', color='danger', url='http://google.com'),
            mock.call().send(),
            mock.call(title='Error', message='Error message', color='danger', url='http://google.com'),
            mock.call().send()
        ])
        log_mock.log.assert_has_calls([
            mock.call(40, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=3),
            mock.call(40, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=3),
            mock.call(40, 'Slack message: Error -> Error message', extra={'labels': {}}, stacklevel=3)
        ])


class LoggerFormatterTestCase(TestCase):

    def setUp(self) -> None:
        self.formatter = log_module.Formatter()

    def test_get_default_dict(self):
        self.assertDictEqual(
            self.formatter._get_default_dict(message='test', severity='INFO'),
            {'message': 'test', 'severity': 'INFO'}
        )

    def test_get_default_extra_dict(self):
        self.assertDictEqual(
            self.formatter._get_default_extra_dict(
                name='everysk-log-formatter-test',
                headers={'X-Request-ID': '123'},
                payload={'data': '123'},
                response={'result': 'ok'},
                traceback='',
                labels={'data': '123'}
            ),
            {
                'logName': 'everysk-log-formatter-test',
                'labels': {'data': '123'},
                'traceback': '',
                'http': {
                    'headers': {'X-Request-ID': '123'},
                    'payload': {'data': '123'},
                    'response': {'result': 'ok'}
                }
            }
        )

    def test_get_default_extra_dict_traceback(self):
        try:
            a = 1 / 0 # pylint: disable=unused-variable
        except ZeroDivisionError:
            self.assertDictEqual(
                self.formatter._get_default_extra_dict(
                    name='everysk-log-formatter-test',
                    headers={'X-Request-ID': '123'},
                    payload={'data': '123'},
                    response={'result': 'ok'},
                    traceback=traceback.format_exc(),
                    labels={'data': '123'}
                ),
                {
                    'logName': 'everysk-log-formatter-test',
                    'labels': {'data': '123'},
                    'traceback': traceback.format_exc(),
                    'http': {
                        'headers': {'X-Request-ID': '123'},
                        'payload': {'data': '123'},
                        'response': {'result': 'ok'},
                    }
                }
            )

    def test_get_default_gcp_dict(self):
        self.assertDictEqual(
            self.formatter._get_default_gcp_dict(headers={'X-Request-ID': '123'}, filename='test.py', line=10, func_name='test_func'),
            {
                'logging.googleapis.com/sourceLocation': {'file': 'test.py', 'function': 'test_func', 'line': 10},
                'logging.googleapis.com/spanId': '',
                'logging.googleapis.com/trace': '',
                'logging.googleapis.com/trace_sampled': False
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_default_gcp_dict_traceparent(self):
        self.assertDictEqual(
            self.formatter._get_default_gcp_dict(
                headers={'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'},
                filename='test.py',
                line=10,
                func_name='test_func'
            ),
            {
                'logging.googleapis.com/sourceLocation': {'file': 'test.py', 'function': 'test_func', 'line': 10},
                'logging.googleapis.com/spanId': '1c6c592f9e46e3fb',
                'logging.googleapis.com/trace': 'projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df',
                'logging.googleapis.com/trace_sampled': True
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_default_gcp_dict_x_cloud_trace_context(self):
        self.assertDictEqual(
            self.formatter._get_default_gcp_dict(
                headers={'x-cloud-trace-context': '4bfa9e049143840bef864a7859f2e5df/2048109991600514043;o=1'},
                filename='test.py',
                line=10,
                func_name='test_func'
            ),
            {
                'logging.googleapis.com/sourceLocation': {'file': 'test.py', 'function': 'test_func', 'line': 10},
                'logging.googleapis.com/spanId': '2048109991600514043',
                'logging.googleapis.com/trace': 'projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df',
                'logging.googleapis.com/trace_sampled': True
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_result_dict(self):
        record = logging.LogRecord(
            args=(),
            exc_info=None,
            func='test_func',
            level=logging.INFO,
            lineno=10,
            msg='test',
            name='everysk-log-formatter-test',
            pathname='test.py',
        )

        record.http_headers = {'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'}
        record.http_payload = {'data': '123'}
        record.http_response = {'result': 'ok'}
        self.assertDictEqual(
            self.formatter._get_result_dict(record),
            {
                'message': 'test',
                'severity': 'INFO',
                'labels': {},
                'logName': 'everysk-log-formatter-test',
                'traceback': '',
                'http': {
                    'headers': {'traceparent': '00-4bfa9e049143840bef864a7859f2e5df-1c6c592f9e46e3fb-01'},
                    'payload': {'data': '123'},
                    'response': {'result': 'ok'},
                },
                'logging.googleapis.com/sourceLocation': {'file': 'test.py', 'function': 'test_func', 'line': 10},
                'logging.googleapis.com/spanId': '1c6c592f9e46e3fb',
                'logging.googleapis.com/trace': 'projects/g-project/traces/4bfa9e049143840bef864a7859f2e5df',
                'logging.googleapis.com/trace_sampled': True
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_get_result_dict_no_http(self):
        record = logging.LogRecord(
            args=(),
            exc_info=None,
            func='test_func',
            level=logging.INFO,
            lineno=10,
            msg='test',
            name='everysk-log-formatter-test',
            pathname='test.py',
        )

        self.assertDictEqual(
            self.formatter._get_result_dict(record),
            {
                'message': 'test',
                'severity': 'INFO',
                'labels': {},
                'logName': 'everysk-log-formatter-test',
                'traceback': '',
                'http': {
                    'headers': {},
                    'payload': {},
                    'response': {},
                },
                'logging.googleapis.com/sourceLocation': {'file': 'test.py', 'function': 'test_func', 'line': 10},
                'logging.googleapis.com/spanId': '',
                'logging.googleapis.com/trace': '',
                'logging.googleapis.com/trace_sampled': False
            }
        )

    @mock.patch.dict('os.environ', {'EVERYSK_GOOGLE_CLOUD_PROJECT': 'g-project'})
    def test_format_message(self):
        record = logging.LogRecord(
            args=(),
            exc_info=None,
            func='test_func',
            level=logging.INFO,
            lineno=10,
            msg='test',
            name='everysk-log-formatter-test',
            pathname='test.py',
        )
        result = self.formatter._get_result_dict(record)
        self.assertEqual(
            self.formatter.formatMessage(record),
            json.dumps(result)
        )


@mock.patch('time.time', mock.MagicMock(return_value=1704067200.0))
class LoggerStdoutTestCase(TestCase):

    def setUp(self) -> None:
        self.log = log_module.Logger(name='everysk-log-stdout-test')

    def test_critical_with_labels(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.critical('with labels', extra={'labels': {}})
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - CRITICAL - {} - with labels\n')

    def test_critical_without_labels(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.critical('without labels')
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - CRITICAL - {} - without labels\n')

    def test_debug_with_labels(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.debug('with labels', extra={'labels': {}})
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - DEBUG - {} - with labels\n')

    def test_debug_without_labels(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.debug('without labels')
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - DEBUG - {} - without labels\n')

    def test_error_with_labels(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.error('with labels', extra={'labels': {}})
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - ERROR - {} - with labels\n')

    def test_error_without_labels(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.error('without labels')
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - ERROR - {} - without labels\n')

    def test_info_with_http_payload_in_bytes(self):
        with mock.patch.object(self.log._log.handler, 'stream') as stream:
            self.log.info('with http payload in bytes', extra={'http_payload': b'{"key": "value"}'})
            stream.write.assert_called_once_with('2024-01-01 00:00:00,000 - INFO - {} - with http payload in bytes\n')


@mock.patch.dict('os.environ', {'LOGGING_JSON': '1'})
class LoggerStackLevelTestCase(TestCase):

    def setUp(self):
        self.default_stacklevel = log_module.Logger._default_stacklevel

    def tearDown(self):
        log_module.Logger._default_stacklevel = self.default_stacklevel

    def test_default(self):
        log = log_module.Logger(name='everysk-log-stacklevel-test')
        with mock.patch.object(log._log.handler, 'stream') as stream:
            log.info('test')
            result = json.loads(stream.write.call_args[0][0])
            self.assertDictEqual(
                result['logging.googleapis.com/sourceLocation'],
                {'file': '/var/app/src/everysk/core/_tests/log.py', 'line': 943, 'function': 'test_default'}
            )

    def test_init(self):
        log_module.Logger._default_stacklevel = 2
        log = log_module.Logger(name='everysk-log-stacklevel-test')
        with mock.patch.object(log._log.handler, 'stream') as stream:
            log.info('test')
            result = json.loads(stream.write.call_args[0][0])
            self.assertDictEqual(
                result['logging.googleapis.com/sourceLocation'],
                {'file': '/var/app/src/everysk/core/log.py', 'line': 662, 'function': 'info'}
            )

        log = log_module.Logger(name='everysk-log-stacklevel-test', stacklevel=3)
        with mock.patch.object(log._log.handler, 'stream') as stream:
            log.info('test')
            result = json.loads(stream.write.call_args[0][0])
            self.assertDictEqual(
                result['logging.googleapis.com/sourceLocation'],
                {'file': '/var/app/src/everysk/core/_tests/log.py', 'line': 963, 'function': 'test_init'}
            )

    def test_class_atribute(self):
        log_module.Logger._default_stacklevel = 2
        log = log_module.Logger(name='everysk-log-stacklevel-test')
        with mock.patch.object(log._log.handler, 'stream') as stream:
            log.info('test')
            result = json.loads(stream.write.call_args[0][0])
            self.assertDictEqual(
                result['logging.googleapis.com/sourceLocation'],
                {'file': '/var/app/src/everysk/core/log.py', 'line': 662, 'function': 'info'}
            )

        class FakeLogger(log_module.Logger):
            stacklevel = 3
        log = FakeLogger(name='everysk-log-stacklevel-test')
        with mock.patch.object(log._log.handler, 'stream') as stream:
            log.info('test')
            result = json.loads(stream.write.call_args[0][0])
            self.assertDictEqual(
                result['logging.googleapis.com/sourceLocation'],
                {'file': '/var/app/src/everysk/core/_tests/log.py', 'line': 985, 'function': 'test_class_atribute'}
            )

    def test_method(self):
        log_module.Logger._default_stacklevel = 2
        log = log_module.Logger(name='everysk-log-stacklevel-test')
        with mock.patch.object(log._log.handler, 'stream') as stream:
            log.info('test')
            result = json.loads(stream.write.call_args[0][0])
            self.assertDictEqual(
                result['logging.googleapis.com/sourceLocation'],
                {'file': '/var/app/src/everysk/core/log.py', 'line': 662, 'function': 'info'}
            )

        log = log_module.Logger(name='everysk-log-stacklevel-test')
        with mock.patch.object(log._log.handler, 'stream') as stream:
            log.info('test', stacklevel=3)
            result = json.loads(stream.write.call_args[0][0])
            self.assertDictEqual(
                result['logging.googleapis.com/sourceLocation'],
                {'file': '/var/app/src/everysk/core/_tests/log.py', 'line': 1005, 'function': 'test_method'}
            )



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import Any


###############################################################################
#   Public Functions Implementation
###############################################################################
def is_float_convertible(value: Any) -> bool:
    """
    Check if a value can be successfully converted to a floating-point number.

    Args:
        value: The value to check for float convertibility.

    Returns:
        bool: True if the value can be converted to a float, False otherwise.

    Example:
        >>> is_float_convertible("3.14")
        True

        >>> is_float_convertible("not_a_number")
        False
    """
    try:
        float(value)
        return True
    except Exception: # pylint: disable=broad-exception-caught
        return False



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.server.applications import create_application
from everysk.server.endpoints import JSONEndpoint
from everysk.server.routing import Route, RouteLazy


###############################################################################
#   TestPublicEndpoint Class Implementation
###############################################################################
class TestPublicEndpoint(JSONEndpoint):
    rest_key_name: str = None
    rest_key_value: str = None

    async def get(self):
        return {'message': 'Hello, World!'}


###############################################################################
#   TestPrivateEndpoint Class Implementation
###############################################################################
class TestPrivateEndpoint(JSONEndpoint):
    rest_key_name: str = 'X-Api-Key'
    rest_key_value: str = '123456'

    async def get(self):
        return {'message': 'Hello, World Private!'}


###############################################################################
#   Application Implementation
###############################################################################
routes = [
    RouteLazy(path='/', endpoint='everysk.server.example_api.TestPublicEndpoint'),
    Route(path='/private', endpoint=TestPrivateEndpoint)
]

app = create_application(routes=routes)


###############################################################################
#   How to run this example
###############################################################################
## To run this example, execute the following command:
# ./run.sh starlette everysk.server.example_api:app

## To test using requests:
# ./run.sh shell
#
# In [1]: import requests
#
# In [2]: requests.get('http://127.0.0.1:8000')
# Out[2]: <Response [200]>
#
# In [3]: requests.get('http://127.0.0.1:8000').content
# Out[3]: b'{"message":"Hello, World!"}'
#
# In [4]: requests.get('http://127.0.0.1:8000/private').content
# Out[4]: b'{"error":"401 -> Unauthorized access to this resource.","code":401,"trace_id":""}'
#
# In [5]: requests.get('http://127.0.0.1:8000/private', headers={'X-Api-Key': '123456'}).content
# Out[5]: b'{"message":"Hello, World Private!"}'
